import tensorflow as tf
from tensorflow import keras
from sklearn.model_selection import KFold, StratifiedKFold
from typing import List, Union
import numpy as np

INT_OR_FLOAT = Union[int, float]

# NOTE: This is a very simple implementation of Bagging.
# And mostly is for demonstration purpose.
# Will make it efficient and better soon enough.
# Also, currently there's just Bagging class, but my plan is to add
# BaggingClassifier and BaggingRegressor like sklearn.


class Bagging:
    def __int__(self,
                build_keras_model_func = None,
                num_estimators: int = 5,
                max_samples: INT_OR_FLOAT = 1.0,
                max_features: INT_OR_FLOAT = 1.0,
                bootstrap: bool = True,
                oob_score: bool = True):
        """
        Args:
             build_keras_model_func: A function that builds and returns Keras model.
        """
        self.build_keras_model_func = build_keras_model_func
        self.num_estimators = num_estimators
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.trained_models = []
        self.all_predictions = []

    def fit(self, X, y):
        for i in range(self.num_estimators):
            bootstrap_indices = np.random.randint(0, len(X), size=len(y))
            X_bootstrap = X[bootstrap_indices]
            y_bootstrap = y[bootstrap_indices]
            model = self.build_keras_model_func()
            model.fit(X_bootstrap, y_bootstrap)
            self.trained_models.append(model)

    def predict(self, X_test):
        for model in self.trained_models:
            preds = model.predict(X_test)
            self.all_predictions.append(preds)
        return np.mean(self.all_predictions, axis=-1)


from tensorflow import keras
from teras.utils import get_normalization_layer
from teras.utils.types import (NormalizationType,
                               ActivationType)


@keras.saving.register_keras_serializable(package="teras.layers.rtdl_resnet")
class RTDLResNetBlock(keras.layers.Layer):
    """
    The ResNet block proposed by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        units: ``int``, default 64,
            Dimensionality of the hidden layer in the ResNet block.

        dropout_hidden: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the hidden dense layer.

        dropout_out: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the output dense layer.

        activation_hidden: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the hidden layer.

        activation_out: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the output layer.

        normalization: ``str`` or ``keras.layers.Layer``, default "batch",
            Normalization layer to normalize the inputs to the RestNet block.

        use_skip_connection: ``bool``, default True,
            Whether to use the skip connection.
    """
    def __init__(self,
                 units: int = 64,
                 dropout_hidden: float = 0.,
                 dropout_out: float = 0.,
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = "relu",
                 normalization: NormalizationType = "batch",
                 use_skip_connection: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.dropout_hidden = dropout_hidden
        self.dropout_out = dropout_out
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out
        self.normalization = normalization
        self.use_skip_connection = use_skip_connection

        if self.normalization is not None:
            self.norm = get_normalization_layer(self.normalization)
        self.dense_hidden = keras.layers.Dense(units=self.units,
                                               activation=self.activation_hidden,
                                               name="resnet_block_dense_hidden")
        self.dropout_hidden = keras.layers.Dropout(self.dropout_hidden,
                                                   name="resnet_block_dropout_hidden")
        self.dropout_out = keras.layers.Dropout(self.dropout_out,
                                                name="resnet_block_dropout_out")
        self.add = keras.layers.Add(name="resnet_block_add")
        self.dense_out = None

    def build(self, input_shape):
        input_dim = input_shape[1]
        self.dense_out = keras.layers.Dense(units=input_dim,
                                            activation=self.activation_out,
                                            name="resnet_block_dense_out")

    def call(self, inputs):
        residual = inputs
        x = inputs
        if self.normalization is not None:
            x = self.norm(x)
        x = self.dense_hidden(x)
        x = self.dropout_hidden(x)
        x = self.dense_out(x)
        x = self.dropout_out(x)
        if self.use_skip_connection:
            x = self.add([x, residual])
        return x

    def get_config(self):
        config = super().get_config()

        if isinstance(self.activation_hidden, str):
            activation_hidden_serialized = self.activation_hidden
        else:
            activation_hidden_serialized = keras.layers.serialize(self.activation_hidden)

        if isinstance(self.activation_out, str):
            activation_out_serialized = self.activation_out
        else:
            activation_out_serialized = keras.layers.serialize(self.activation_out)

        if isinstance(self.normalization, str):
            normalization_serialized = self.normalization
        else:
            normalization_serialized = keras.layers.serialize(self.normalization)

        config.update({'units': self.units,
                       'dropout_hidden': self.dropout_hidden,
                       'dropout_out': self.dropout_out,
                       'activation_hidden': activation_hidden_serialized,
                       'activation_out': activation_out_serialized,
                       'normalization': normalization_serialized,
                       'use_skip_connection': self.use_skip_connection,
                       })
        return config



import tensorflow as tf


@tf.function
def vime_mask_generator(x,
                        p_m: float = 0.3):
    """
    Generates mask vector for self and semi-supervised learning
    Args:
        x:
            Dataset.

        p_m: ``float``, default 0.3,
            Corruption probability

    Returns:
        mask: binary mask matrix

    """
    mask = tf.random.stateless_binomial(shape=tf.shape(x),
                                        seed=(0, 0),
                                        counts=1,
                                        probs=p_m,
                                        output_dtype=tf.float32)
    return mask


@tf.function
def vime_pretext_generator(x, mask):
    """
    Generates corrupted samples for self and semi-supervised learning

    Args:
        x:
            Dataset.

        mask:
            Mask matrix

    Returns:
        mask_corrupted: Final mask matrix after corruption
        x_tilde: Corrupted feature matrix
    """
    x = tf.cast(x, dtype=tf.float32)
    num_samples = tf.shape(x)[0]
    dim = tf.shape(x)[1]

    x_bar = tf.TensorArray(size=dim, dtype=tf.float32)
    for i in range(dim):
        idx = tf.random.shuffle(tf.range(num_samples))
        x_bar = x_bar.write(i, tf.gather(x[:, i], idx))
    x_bar = tf.transpose(x_bar.stack())

    # Corrupt Samples
    x_tilde = x * (1 - mask) + x_bar * mask

    m_new = (x != x_tilde)
    m_new = tf.cast(m_new, dtype="float32")
    return m_new, x_tilde


def preprocess_input_vime_semi(x_labeled,
                               y_labeled,
                               x_unlabeled,
                               batch_size=None):
    """
    VIME's semi supervised training requires a labeled and an unlabeled training set.
    It also allows user to specify the unlabeled/labeled split.
    Hence, the labeled and unlabeled dataset sizes will almost always differ.
    And here's the issue, Keras's default fit method will raise incompatible input cardinalities error
    if we try to pass labeled and unlabeled datasets to `x` parameter as a dict and pass labeled targets to `y`
    To allow for varying sizes for labeled and unlabeled dataset, this function comes to clutch.
    It fuses together the labeled and unlabeled dataset and makes sure that each labeled batch of data is accompanied
    by an unlabeled batch of data as it is required by the VIME architecture.

    Args:
        x_labeled: Labeled training set
        y_labeled: Targets array for labeled training set
        x_unlabeled: Unlabeled dataset
        batch_size: Batch size
    """
    # The official implementation uses random batches at each step, we'll do the same
    num_samples_labeled = tf.cast(tf.shape(x_labeled)[0], tf.int64)
    num_samples_unlabeled = tf.cast(tf.shape(x_unlabeled)[0], tf.int64)

    x_l_ds = (tf.data.Dataset
              .from_tensor_slices({"x_labeled": x_labeled, "y_labeled": y_labeled}, name="labeled_dataset")
              .shuffle(buffer_size=num_samples_labeled, reshuffle_each_iteration=True)
              )

    x_u_ds = (tf.data.Dataset
              .from_tensor_slices({"x_unlabeled": x_unlabeled}, "unlabeled_dataset")
              .shuffle(buffer_size=num_samples_unlabeled, reshuffle_each_iteration=True)
              )

    # If there are fewer unlabeled samples than labeled samples, we'll repeat unlabeled dataset
    # because it is essential to have a batch of unlabeled dataset along with labeled dataset
    # for semi supervised training in VIME.
    # In case when unlabeled dataset contains more samples than labeled, we won't alter anything
    # Because by default, tensorflow dataset will produce batches equal to math.ceil(min(len(dataset_1), len(dataset_2)) / batch_size)
    if num_samples_labeled > num_samples_unlabeled:
        num_times_to_repeat = tf.cast(tf.math.ceil(20/3), dtype=tf.int64)
        x_u_ds = x_u_ds.repeat(num_times_to_repeat)

    x_merged = tf.data.Dataset.zip((x_l_ds, x_u_ds), "merged_dataset")
    if batch_size is not None:
        x_merged = x_merged.batch(batch_size=batch_size)

    return x_merged


def preprocess_input_vime_self(x_unlabeled,
                               p_m=0.3,
                               m_labeled=None,
                               x_tilde=None,
                               batch_size=None
                               ):
    """
    Generates mask and corrupted samples if they aren't specified by the user
    and fuses them together with the unlabeled dataset.

    Args:
        x_unlabeled: Unlabeled dataset
        p_m: Corruption Probability
        m_labeled: Labeled mask generated by pretext_generator
        x_tilde: Corrupted dataset
        batch_size: Batch size

    Returns:
        TensorFlow dataset
    """

    if m_labeled is None and x_tilde is None:
        m_unlabeled = vime_mask_generator(p_m,
                                          x_unlabeled)
        m_labeled, x_tilde = vime_pretext_generator(m_unlabeled, x_unlabeled)

    dataset = tf.data.Dataset.from_tensor_slices((x_tilde,
                                                  {"mask_estimator": m_labeled, "feature_estimator": x_unlabeled}))

    if batch_size is not None:
        dataset = dataset.batch(batch_size=batch_size)

    return dataset


import tensorflow as tf
from teras.utils import tf_random_choice


# Soft binary gates
@tf.function
def and_operator(x, d):
    out = tf.add(x, -d + 1.5)
    out = tf.tanh(out)
    return out


@tf.function
def or_operator(x, d):
    out = tf.add(x, d - 1.5)
    out = tf.tanh(out)
    return out


# Feature Selection
@tf.function
def binary_threshold(x, eps=0.1):
    x = tf.abs(x) - eps
    return 0.5 * binary_activation(x) + 0.5


@tf.function
def binary_activation(x):
    forward = tf.sign(x)
    backward = tf.tanh(x)
    return backward + tf.stop_gradient(forward - backward)


@tf.function
def generate_random_mask(input_dim,
                         keep_feature_prob_arr=None,
                         num_literals_per_formula_arr=None,
                         num_formulas=None):
    literals_random_mask = tf.TensorArray(size=0, dynamic_size=True, dtype=tf.float32)
    formulas_random_mask = tf.TensorArray(size=num_formulas, dtype=tf.float32)

    arr_literal_index = 0
    p_index = 0
    for i in range(num_formulas):
        if i % tf.shape(num_literals_per_formula_arr)[0] == 0 and i != 0:
            p_index = (p_index + 1) % tf.shape(keep_feature_prob_arr)[0]

        p_i = tf.gather(keep_feature_prob_arr, p_index)
        mask = tf_random_choice([1., 0.], input_dim, p=[p_i, 1 - p_i])
        while tf.math.reduce_sum(mask) == 0:
            mask = tf_random_choice([1., 0.], input_dim, p=[p_i, 1 - p_i])
        num_literals_in_formula = tf.gather(num_literals_per_formula_arr, i % tf.shape(num_literals_per_formula_arr)[0])
        formulas_random_mask.write(i, tf.identity(mask))
        for _ in range(num_literals_in_formula):
            literals_random_mask.write(arr_literal_index, tf.identity(mask))
            arr_literal_index += 1

    literals_random_mask = literals_random_mask.stack()
    literals_random_mask = tf.squeeze(literals_random_mask, axis=1)
    formulas_random_mask = formulas_random_mask.stack()
    formulas_random_mask = tf.squeeze(formulas_random_mask, axis=1)
    return tf.transpose(literals_random_mask), \
           tf.transpose(formulas_random_mask)


@tf.function
def extension_matrix(v,
                     num_formulas,
                     num_literals_per_formula_arr):
    formula_index = 0
    mat = tf.TensorArray(size=0, dynamic_size=True, dtype=tf.float32)
    arr_mat_index = 0
    for i in range(num_formulas):
        n_nodes_in_tree = tf.gather(num_literals_per_formula_arr, i % tf.shape(num_literals_per_formula_arr)[0])
        v[formula_index].assign(1.)
        for _ in range(n_nodes_in_tree):
            mat.write(arr_mat_index, tf.identity(v))
            arr_mat_index += 1
        formula_index += 1
        v[:].assign(0.)
    mat = mat.stack()
    return dense_ndim_array_to_sparse_tensor(tf.transpose(mat))


# DNF Structure
@tf.function
def create_conjunctions_indicator_matrix(b,
                                         total_number_of_conjunctions,
                                         conjunctions_depth_arr,
                                         ):
    n_different_depth = tf.shape(conjunctions_depth_arr)[0]
    n_literals_in_group = tf.reduce_sum(conjunctions_depth_arr)
    array_size = (total_number_of_conjunctions // n_different_depth) * tf.shape(conjunctions_depth_arr)[0]
    result = tf.TensorArray(size=array_size, dtype=tf.bool)
    arr_index = 0
    for i in range(total_number_of_conjunctions // n_different_depth):
        s = 0
        for d in conjunctions_depth_arr:
            b[i * n_literals_in_group + s: i * n_literals_in_group + s + d].assign(True)
            s += d
            result.write(arr_index, tf.identity(b))
            arr_index += 1
            b[:].assign(False)
    result = result.stack()
    return tf.transpose(result)


@tf.function
def create_formulas_indicator_matrix(c,
                                     num_formulas,
                                     num_conjunctions_arr,
                                     ):
    result = tf.TensorArray(size=num_formulas, dtype=tf.bool)
    base = 0
    for i in range(num_formulas):
        n_conjunctions = tf.gather(num_conjunctions_arr, i % tf.shape(num_conjunctions_arr)[0])
        c[base: base + n_conjunctions].assign(True)
        result.write(i, tf.identity(c))
        c[:].assign(False)
        base += n_conjunctions
    result = result.stack()
    return tf.transpose(result)


@tf.function
def compute_total_number_of_literals(num_formulas,
                                     num_conjunctions_arr,
                                     conjunctions_depth_arr):
    return (tf.reduce_sum(num_conjunctions_arr) * tf.reduce_sum(conjunctions_depth_arr) * num_formulas) // (
                tf.shape(conjunctions_depth_arr)[0] * tf.shape(num_conjunctions_arr)[0])


@tf.function
def compute_total_number_of_conjunctions(num_formulas, num_conjunctions_arr):
    return (num_formulas // tf.shape(num_conjunctions_arr)[0]) * tf.reduce_sum(num_conjunctions_arr)


@tf.function
def compute_num_literals_per_formula(num_conjunctions_arr,
                                   conjunctions_depth_arr):
    num_literals_per_formula_arr = []
    for n_conjunctions in num_conjunctions_arr:
        num_literals_per_formula_arr.append(
            (n_conjunctions // tf.shape(conjunctions_depth_arr)[0]) * tf.reduce_sum(conjunctions_depth_arr))
    return num_literals_per_formula_arr


# General Functions
@tf.function
def dense_ndim_array_to_sparse_tensor(arr):
    if arr.dtype == tf.bool:
        idx = tf.where(tf.not_equal(arr, False))
    else:
        idx = tf.where(tf.not_equal(arr, 0.0))
    idx = tf.cast(idx, dtype=tf.int64)
    if arr.dtype == tf.bool:
        return tf.SparseTensor(idx,
                               tf.ones(tf.shape(idx)[0]),
                               tf.cast(tf.shape(arr), dtype=tf.int64)
                               )
    else:
        return tf.SparseTensor(idx,
                               tf.gather_nd(arr, idx),
                               tf.cast(tf.shape(arr), dtype=tf.int64)
                               )


@tf.function
def tf_matmul_sparse_dense(sparse_A,
                           dense_B):
    return tf.transpose(tf.sparse.sparse_dense_matmul(tf.sparse.transpose(sparse_A), tf.transpose(dense_B)))




# NODE Utility Function(s)
import tensorflow as tf


def sparsemoid(x):
    """
    Sparsemoid function as implemented by the authors of
    Neural Oblivious Decision Tree Ensembles (NODE) paper.
    It is used as a bin function in the NODE architecture.

    Reference:
        https://github.com/Qwicen/node/blob/master/lib/nn_utils.py
    """
    return tf.clip_by_value(0.5 * x + 0.5, 0., 1.)


from tensorflow import keras
from teras.layers.tabnet.tabnet_feature_transformer_block import TabNetFeatureTransformerBlock


@keras.saving.register_keras_serializable("teras.layers.tabnet")
class TabNetFeatureTransformer(keras.layers.Layer):
    """
    Feature Transformer as proposed by Sercan et al. in TabNet paper.
    It is made up of ``FeatureTransformerBlock`` building blocks.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        units: ``int``, default 32,
            the dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a ``2 * feature_transformer_dim``
            output and half of it is used to determine the
            non-linearity of the GLU activation where the other half is used as an
            input to GLU, and eventually ``feature_transformer_dim`` output is
            transferred to the next layer.

        num_shared_layers: ``int``, default 2.
            Number of shared layers to use in the Feature Transformer.
            These shared layers are `shared` across decision steps.

        num_decision_dependent_layers: ``int``, default 2.
            Number of decision dependent layers to use.
            In simple words, ``num_decision_dependent_layers`` are created for each
            decision step in the ``num_decision_steps``.
            For instance, if ``num_decision_steps = 5`` and  ``num_decision_dependent_layers = `2`
            then 10 layers will be created, 2 for each decision step.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: ``int``, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the `batch_size` used for training.
            And most importantly, the ``batch_size`` must always be divisible by the ``virtual_batch_size``.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the layer, every other layer utilizes
            normalized residuals, where ``residual_normalization_factor`` determines the scale of normalization.
    """
    shared_layers = None

    def __init__(self,
                 units: int = 32,
                 num_shared_layers: int = 2,
                 num_decision_dependent_layers: int = 2,
                 batch_momentum: float = 0.9,
                 virtual_batch_size: int = 64,
                 residual_normalization_factor: float = 0.5,
                 **kwargs):

        if num_shared_layers == 0 and num_decision_dependent_layers == 0:
            raise ValueError("Feature Transformer requires at least one of either shared or decision depenedent layers."
                             " But both `num_shared_layers` and `num_decision_dependent_layers` were passed a 0 value.")

        super().__init__(**kwargs)
        self.units = units
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.batch_momentum = batch_momentum
        self.virtual_batch_size = virtual_batch_size
        self.residual_normalization_factor = residual_normalization_factor

        # Since shared layers should be accessible across instances so we make them as class attributes
        # and use class method to instantiate them, the first time an instance of this FeatureTransformer
        # class is made.
        if self.shared_layers is None and num_shared_layers > 0:
            self._create_shared_layers(num_shared_layers,
                                       units,
                                       batch_momentum=batch_momentum,
                                       virtual_batch_size=virtual_batch_size,
                                       residual_normalization_factor=residual_normalization_factor)

        self.inner_block = keras.models.Sequential(name="inner_block")

        self.inner_block.add(self.shared_layers)

        if self.num_decision_dependent_layers > 0:
            decision_dependent_layers = keras.models.Sequential(name=f"decision_dependent_layers")
            for j in range(self.num_decision_dependent_layers):
                if j == 0 and self.num_shared_layers == 0:
                    # then it means that this is the very first layer in the Feature Transformer
                    # because no shard layers exist.
                    # hence we shouldn't use residual normalization
                    use_residual_normalization = False
                else:
                    use_residual_normalization = True
                decision_dependent_layers.add(TabNetFeatureTransformerBlock(
                    units=self.units,
                    batch_momentum=self.batch_momentum,
                    virtual_batch_size=self.virtual_batch_size,
                    use_residual_normalization=use_residual_normalization,
                    name=f"decision_dependent_layer_{j}")
                )

            self.inner_block.add(decision_dependent_layers)

    @classmethod
    def _create_shared_layers(cls, num_shared_layers, units, **kwargs):
        # The outputs of first layer in feature transformer isn't residual normalized
        # so we use a pointer to know which layer is first.
        # Important to keep in mind that shared layers ALWAYS precede the decision dependent layers
        # BUT we want to allow user to be able to use NO shared layers at all, i.e. 0.
        # Hence, only if the `num_shared_layers` is 0, should the first dependent layer for each step
        # be treated differently.
        # Initialize the block of shared layers
        cls.shared_layers = keras.models.Sequential(name="shared_layers")
        for i in range(num_shared_layers):
            if i == 0:
                # First layer should not use residual normalization
                use_residual_normalization = False
            else:
                use_residual_normalization = True
            cls.shared_layers.add(TabNetFeatureTransformerBlock(units=units,
                                                                use_residual_normalization=use_residual_normalization,
                                                                name=f"shared_layer_{i}",
                                                                **kwargs))

    @classmethod
    def reset_shared_layers(cls):
        """
        Resets the shared layers.

        You must reset shared layers, when you want to create
        new instances of ``TabNetFeatureTransformer`` for ``TabNetDecoder``
        after you have created those for the ``TabNetEncoder``.
        """
        cls.shared_layers = None

    def call(self, inputs):
        outputs = self.inner_block(inputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({'units': self.units,
                       'num_shared_layers': self.num_shared_layers,
                       'num_decision_dependent_layers': self.num_decision_dependent_layers,
                       'batch_momentum': self.batch_momentum,
                       'virtual_batch_size': self.virtual_batch_size,
                       'residual_normalization_factor': self.residual_normalization_factor,
                       })
        return config



import tensorflow as tf
from tensorflow import keras
from teras.layerflow.models.tvae import TVAE as _TVAE_LF
from teras.utils.types import UnitsValuesType


@keras.saving.register_keras_serializable(package="teras.models")
class TVAEEncoder(keras.Model):
    """
    Encoder for the TVAE model as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        latent_dim: ``int``, default 128.
            Dimensionality of the learned latent space

        units_values: ``List[int]`` or ``Tuple[int]``, default (128, 128),
            A list or tuple of integers.
            For each value in the sequence, a dense layer of that
            dimensions (units) is added to construct a compression block.
    """
    def __init__(self,
                 latent_dim: int = 128,
                 units_values: UnitsValuesType = (128, 128),
                 **kwargs):
        super().__init__(**kwargs)

        if not isinstance(units_values, (list, tuple)):
            raise ValueError(f"""`units_values` must be a list or tuple of units which determines
                        the number of compression layers and the dimensionality of those layers.
                        Received: {units_values}""")

        self.latent_dim = latent_dim
        self.units_values = units_values

        self.compression_block = keras.models.Sequential(name="compression_block")
        for i, units in enumerate(self.units_values, start=1):
            self.compression_block.add(keras.layers.Dense(units=units,
                                                          activation="relu",
                                                          name=f"compression_layer_{i}"))
        self.dense_mean = keras.layers.Dense(units=self.latent_dim,
                                             name="mean")
        self.dense_log_var = keras.layers.Dense(units=self.latent_dim,
                                                name="log_var")

    def call(self, inputs):
        h = self.compression_block(inputs)
        mean = self.dense_mean(h)
        log_var = self.dense_log_var(h)
        std = tf.exp(0.5 * log_var)
        return mean, std, log_var

    def get_config(self):
        config = super().get_config()
        new_config = {'latent_dim': self.latent_dim,
                      'units_values': self.units_values
                      }
        config.update(new_config)
        return config


@keras.saving.register_keras_serializable(package="teras.models")
class TVAEDecoder(keras.Model):
    """
    Encoder for the TVAE model as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        data_dim: `int`,
            Dimensionality of the input dataset. It is used to create the last layer for
            the decompression_block that projects data back to the original dimensions.
        units_values: `List[int]` or `Tuple[int]`,
            A list or tuple of integers. For each value in the sequence,
            a dense layer of that dimensionality is added to construct a decompression block.
            Note that, a dense layer of `data_dim` is also appended at the of decompression
            block to project data back to the original data dimensions.
    """
    def __init__(self,
                 data_dim: int = None,
                 units_values: UnitsValuesType = (128, 128),
                 **kwargs):
        super().__init__(**kwargs)

        if data_dim is None:
            raise ValueError("`data_dim` cannot be None, "
                             "as `data_dim` value is required to project data back to original data dimensions.")

        if not isinstance(units_values, (list, tuple)):
            raise ValueError(f"""`units_values` must be a list or tuple of units which determines
                        the number of decompression layers and the dimensionality of those layers.
                        Received: {units_values}""")

        self.data_dim = data_dim
        self.units_values = units_values

        self.decompression_block = keras.models.Sequential(name="decompression_block")
        for i, units in enumerate(self.units_values, start=1):
            self.decompression_block.add(keras.layers.Dense(units=units,
                                                            activation="relu",
                                                            name=f"decompression_layer_{i}"
                                                            )
                                         )
        self.decompression_block.add(keras.layers.Dense(units=self.data_dim,
                                                        name="projection_to_data_dim"))

        self.sigmas = tf.Variable(initial_value=keras.initializers.ones()(shape=(self.data_dim,)) * 0.1,
                                  trainable=True,
                                  name="sigmas")

    def call(self, inputs):
        x_generated = self.decompression_block(inputs)
        return x_generated, self.sigmas

    def get_config(self):
        config = super().get_config()
        new_config = {'data_dim': self.data_dim,
                      'units_values': self.units_values
                      }
        config.update(new_config)
        return config


@keras.saving.register_keras_serializable(package="teras.models")
class TVAE(_TVAE_LF):
    """
    TVAE model for tabular data generation
    based on the architecture proposed by the
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        input_dim: ``int``, dimensionality of the input dataset.
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``TVAEDataSampler`` instance used in sampling
            the dataset.

        metadata: ``dict``,
            A dictionary containing metadata for all features in the input data.
            This metadata is computed during the data transformation step and can be accessed
            from ``.get_metadata()`` method of ``TVAEDataTransformer`` instance.

        units_values: ``List[int]`` or ``Tuple[int]``, default (128, 128),
            A list or tuple of units to construct the compression block in ``TVAEEncoder``.
            For each value in the sequence, a dense layer of that
            dimensions (units) is added to construct a compression block.

        decoder_units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units to construct the decompression block in ``TVAEDecoder``.
            For each value in the sequence,
            a dense layer of that dimensionality is added to construct a decompression block.
            Note that, a dense layer of ``data_dim`` is also appended at the of decompression
            block to project data back to the original data dimensions.

        latent_dim: `int`, default 128,
            Dimensionality of the learned latent space.

        loss_factor: `float`, default 2,
            Hyperparameter used in the computation of ELBO loss for TVAE.
            It controls how much the cross entropy loss contributes to the overall loss.
            It is directly proportional to the cross entropy loss.
    """
    def __init__(self,
                 input_dim: int,
                 metadata: dict = None,
                 encoder_units_values: UnitsValuesType = (128, 128),
                 decoder_units_values: UnitsValuesType = (128, 128),
                 latent_dim: int = 128,
                 loss_factor: float = 2,
                 **kwargs):
        if metadata is None:
            raise ValueError("`metadata` cannot be None. "
                             "You can access it through `.get_metadata()` method of the `TVAEDataTransformer` instance.")
        encoder = TVAEEncoder(latent_dim=latent_dim,
                              units_values=encoder_units_values)
        decoder = TVAEDecoder(data_dim=input_dim,
                              units_values=decoder_units_values)
        super().__init__(encoder=encoder,
                         decoder=decoder,
                         latent_dim=latent_dim,
                         loss_factor=loss_factor,
                         **kwargs)
        self.input_dim = input_dim
        self.metadata = metadata
        self.encoder_units_values = encoder_units_values
        self.decoder_units_values = decoder_units_values
        self.latent_dim = latent_dim
        self.loss_factor = loss_factor

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'metadata': self.metadata,
                  'encoder_units_values': self.encoder_units_values,
                  'decoder_units_values': self.decoder_units_values,
                  'latent_dim': self.latent_dim,
                  'loss_factor': self.loss_factor,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim,
                   **config)



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K
from teras.layers.numerical_features_extractor import NumericalFeaturesExtractor
from teras.losses.tabnet import reconstruction_loss
import tensorflow_probability as tfp


@keras.saving.register_keras_serializable(package="keras.layerflow.models")
class TabNet(keras.Model):
    """
    TabNet model class with LayerFlow design.

    TabNet is a novel high-performance and interpretable canonical
    deep tabular data learning architecture.
    TabNet uses sequential attention to choose which features to reason
    from at each decision step, enabling interpretability and more
    efficient learning as the learning capacity is used for the most
    salient features.

    TabNet is proposed by Sercan et al. in TabNet paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        categorical_feature_embedding: ``keras.layers.Layer``,
            An instance of ``CategoricalFeatureEmbedding`` layer to embedd categorical features
            or any layer that can work in its place.
            You can import the ``CategoricalFeatureEmbedding`` layer as follows,
                >>> from teras.layers import CategoricalFeatureEmbedding

            Important Note: The embedding layer must have a dimensionality of ``1`` else it will
            result in error.


        encoder: ``keras.layers.Layer``,
            An instance of Encoder layer to encode feature embeddings,
            or any layer that can work in place of Encoder for that purpose.
            If None, an Encoder layer with default values will be used.
                >>> from teras.layers import TabNetEncoder

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.

            REMEMBER: In case you're using this model as a base model for pretraining, you MUST leave
            this argument as None.

            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """
    def __init__(self,
                 input_dim: int,
                 features_metadata: dict,
                 categorical_feature_embedding: keras.layers.Layer = None,
                 encoder: keras.layers.Layer = None,
                 head: keras.layers.Layer = None,
                 **kwargs):
        numerical_features_exist = len(features_metadata["numerical"]) > 0
        categorical_features_exist = len(features_metadata["categorical"]) > 0

        if categorical_feature_embedding is None and categorical_features_exist:
            raise ValueError("`categorical_feature_embedding` cannot be None when categorical features exist "
                             "in the dataset. Please pass an instance of `CategoricalFeatureEmbedding` which you "
                             "can import as follows, \n"
                             "`from teras.layers import CategoricalFeatureEmbedding`")

        inputs = keras.layers.Input(shape=(input_dim,))
        if encoder is None:
            raise ValueError("`encoder` cannot be None. Please pass an instance of `TabNetEncoder` layer. "
                             "You can import it as, `from teras.layers import TabNetEncoder``")

        # Important to note that, this is one model where both numerical and categorical feature embeddings
        # can be None. Because tabnet, by default uses raw numerical features.

        if categorical_feature_embedding is None:
            # then there's only numerical features
            features_embeddings = NumericalFeaturesExtractor(features_metadata)(inputs)
        else:
            # then there are definitely categorical features but might or might not be numerical features
            features_embeddings = categorical_feature_embedding(inputs)
            features_embeddings = keras.layers.Flatten()(features_embeddings)
            if numerical_features_exist:
                numerical_embeddings = NumericalFeaturesExtractor(features_metadata)(inputs)
                features_embeddings = keras.layers.Concatenate(axis=1)([features_embeddings, numerical_embeddings])
        outputs = encoder(features_embeddings)
        if head is not None:
            outputs = head(outputs)
        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)
        self.input_dim = input_dim
        self.features_metadata = features_metadata
        self.categorical_feature_embedding = categorical_feature_embedding
        self.encoder = encoder
        self.head = head

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'features_metadata': self.features_metadata,
                       'categorical_feature_embedding': keras.layers.serialize(self.categorical_feature_embedding),
                       'encoder': keras.layers.serialize(self.encoder),
                       'head': keras.layers.serialize(self.head),
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        features_metadata = config.pop("features_metadata")
        categorical_feature_embedding = keras.layers.deserialize(config.pop("categorical_feature_embedding"))
        encoder = keras.layers.deserialize(config.pop("encoder"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   features_metadata=features_metadata,
                   categorical_feature_embedding=categorical_feature_embedding,
                   encoder=encoder,
                   head=head,
                   **config)


@keras.saving.register_keras_serializable(package="keras.layerflow.models")
class TabNetPretrainer(keras.Model):
    """
    TabNetPretrainer with LayerFlow desing.

    It is an encoder-decoder model based on the TabNet architecture,
    where the TabNet model acts as an encoder while a separate decoder
    is used to reconstruct the input features.

    It is proposed by Sercan et al. in TabNet paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        model: ``TabNet``,
            An instance of base ``TabNet`` model to pretrain.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        decoder: ``keras.layers.Layer``,
            An instance of ``TabNetDecoder`` layer or any custom layer
            that can be used in its place to reconstruct the input
            features from the encoded representations.
            You can import the ``TabNetDecoder`` layer as
                >>> from teras.layers import TabNetDecoder

        missing_feature_probability: ``float``, default 0.3,
            Fraction of features to randomly mask i.e. make them missing.
            Missing features are introduced in the pretraining dataset and
            the probability of missing features is controlled by the parameter.
            The pretraining objective is to predict values for these missing features,
            (pre)training the ``TabNet`` model in the process.
    """
    def __init__(self,
                 model: TabNet,
                 features_metadata: dict,
                 decoder: keras.layers.Layer = None,
                 missing_feature_probability: float = 0.3,
                 **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.features_metadata = features_metadata
        self.decoder = decoder
        self.missing_feature_probability = missing_feature_probability

        self.numerical_features_indices = list(self.features_metadata["numerical"].values())
        self._categorical_features_exist = len(self.features_metadata["categorical"]) > 0
        self._numerical_features_exist = len(self.features_metadata["numerical"]) > 0
        self.input_dim = len(self.features_metadata["categorical"]) + len(self.features_metadata["numerical"])

        self.binary_mask_generator = tfp.distributions.Binomial(total_count=1,
                                                                probs=self.missing_feature_probability,
                                                                name="binary_mask_generator")
        self._reconstruction_loss_tracker = keras.metrics.Mean(name="reconstruction_loss")

    def get_pretrained_model(self):
        """Returns pretrained model"""
        return self.model

    @property
    def pretrained_model(self):
        """Returns pretrained model"""
        return self.model

    @property
    def feature_importances_per_sample(self):
        """Returns feature importances per sample computed during training."""
        return tf.concat(self.encoder.feature_importances_per_sample, axis=0)

    @property
    def feature_importances(self):
        """Returns average feature importances across samples computed during training."""
        return tf.reduce_mean(self.feature_importances_per_sample, axis=0)

    def compile(self,
                loss=reconstruction_loss,
                optimizer=keras.optimizers.Adam(learning_rate=0.001),
                **kwargs):
        super().compile(**kwargs)
        self.reconstruction_loss = loss
        self.optimizer = optimizer

    def call(self, inputs, mask=None):
        # this mask below is what `S` means in the paper, where if an index contains
        # value 1, it means that it is missing
        encoder_input = (1 - mask) * inputs
        # The paper says,
        # The TabNet encoder inputs (1 − S) · f, where f is the original features
        # and the decoder outputs the reconstructed features, S · f^, where f^ is the reconstructed features
        # We initialize P[0] = (1 − S) in encoder so that the model emphasizes merely on the known features.
        # -- So we pass the mask from here, the encoder checks if it received a value for mask, if so it won't
        # initialized the `mask_values` variable in its call method to zeros.
        encoded_representations = self.model.encoder(encoder_input, mask=(1 - mask))
        decoder_outputs = self.decoder(encoded_representations)
        return decoder_outputs

    def train_step(self, data):
        if isinstance(data, tuple):
            data = data[0]
        cat_embed_dim = 0
        num_embed_dim = 0

        with tf.GradientTape() as tape:
            embedded_inputs = None
            if self._categorical_features_exist:
                categorical_embeddings = self.model.categorical_feature_embedding(data)
                categorical_embeddings = K.flatten(categorical_embeddings)
                cat_embed_dim = tf.shape(categorical_embeddings)[1]
                embedded_inputs = categorical_embeddings
            # We need to concatenate numerical features to the categorical embeddings
            if self._numerical_features_exist:
                if (hasattr(self.model, "numerical_feature_embedding")
                        and self.model.numerical_feature_embedding is not None):
                    numerical_embeddings = self.model.numerical_feature_embedding(data)
                else:
                    numerical_embeddings = tf.gather(data,
                                                     indices=self.numerical_features_indices,
                                                     axis=1)
                num_embed_dim = tf.shape(numerical_embeddings)[1]
                if embedded_inputs is None:
                    embedded_inputs = numerical_embeddings
                else:
                    embedded_inputs = K.concatenate([embedded_inputs, numerical_embeddings], axis=1)
            total_dim = cat_embed_dim + num_embed_dim
            embedded_inputs.set_shape((None, total_dim))
            # Generate mask to create missing samples
            batch_size = tf.shape(embedded_inputs)[0]
            mask = self.binary_mask_generator.sample(sample_shape=(batch_size, total_dim))
            tape.watch(mask)
            # Reconstruct samples
            reconstructed_samples = self(embedded_inputs, mask=mask)
            # Compute reconstruction loss
            loss = self.reconstruction_loss(real_samples=embedded_inputs,
                                            reconstructed_samples=reconstructed_samples,
                                            mask=mask)
        gradients = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_weights))
        self._reconstruction_loss_tracker.update_state(loss)
        # If user has passed any additional metrics to compile, we should update their states
        if len(self.compiled_metrics.metrics) > 0:
            self.compiled_metrics.update_state(embedded_inputs, reconstructed_samples)
        # If user has passed any additional losses to compile, we should call them
        if self.compiled_loss._losses is not None:
            self.compiled_loss(embedded_inputs, reconstructed_samples)
        results = {m.name: m.result() for m in self.metrics}
        return results

    def get_config(self):
        config = super().get_config()
        config.update({'model': keras.layers.serialize(self.model),
                       'features_metadata': self.features_metadata,
                       'decoder': keras.layers.serialize(self.decoder),
                       'missing_feature_probability': self.missing_feature_probability
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        features_metadata = config.pop("features_metadata")
        decoder = keras.layers.deserialize(config.pop("decoder"))
        return cls(model=model,
                   features_metadata=features_metadata,
                   decoder=decoder,
                   **config)



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class MultiHeadInterSampleAttention(keras.layers.Layer):
    """
    MultiHeadInterSampleAttention layer as proposed by Gowthami Somepalli et al.
    in the paper SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.
    Unlike the usual MultiHeadAttention layer, this MultiHeadInterSampleAttention layer,
    as the name enunciates, applies attention over samples/rows instead of features/columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        num_heads: `int`, default 8,
            Number of Attention heads to use
        key_dim: `int`, default 32,
            Key dimensionality for attention.
        dropout: `float`, default 0.1,
            Dropout rate to use.
    """
    def __init__(self,
                 num_heads: int = 8,
                 key_dim: int = 32,
                 dropout: float = 0.1,
                 **kwargs):
        super().__init__()
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.dropout = dropout
        self.multi_head_attention = keras.layers.MultiHeadAttention(num_heads=self.num_heads,
                                                                    key_dim=self.key_dim,
                                                                    dropout=dropout,
                                                                    **kwargs)

    def call(self, inputs):
        # Expected inputs shape: (b, n, d)
        # b: batch_size, n: num_features, d: embedding_dim
        x = inputs
        x = tf.reshape(x, shape=(1,
                                 tf.shape(x)[0],
                                 tf.shape(x)[1] * tf.shape(x)[2]))
        x = self.multi_head_attention(x, x)
        x = tf.reshape(x, shape=tf.shape(inputs))
        return x

    def get_config(self):
        config = super().get_config()
        config.update({'num_heads': self.num_heads,
                       'key_dim': self.key_dim,
                       'dropout': self.dropout,
                       })
        return config



import tensorflow as tf
from tensorflow import keras
from teras.utils.dnfnet import (tf_matmul_sparse_dense,
                                generate_random_mask,
                                extension_matrix,
                                binary_threshold)
from teras.utils.types import FloatSequence


@keras.saving.register_keras_serializable(package="teras.layers.dnfnet")
class DNFNetFeatureSelection(keras.layers.Layer):
    """
    DNFNetFeatureSelection layer based on the Feature Selection component
    proposed by Liran Katzir et al.
    in the paper Net-DNF: Effective Deep Modeling of Tabular Data.

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        data_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        num_formulas: ``int``, default 256,
            Number of DNF formulas to use.
            Each DNF formula is analogous to a tree in tree based ensembles.

        keep_feature_prob_arr: ``List[float]`` or ``Tuple[float]``, default [0.1, 0.3, 0.5, 0.7, 0.9],
            Feature probability array.
            It is used  to randomly select a probability value that is used
            in the random selection of input features.

        elastic_net_beta: ``float``, default 0.4,
            Used in the computation of Elastic Net Regularization.

        binary_threshold_eps: ``float``, default 1.0,
            Used in the computation of learnable mask.
    """
    def __init__(self,
                 input_dim: int,
                 num_formulas: int = 256,
                 keep_feature_prob_arr: FloatSequence = [0.1, 0.3, 0.5, 0.7, 0.9],
                 elastic_net_beta: float = 0.4,
                 binary_threshold_eps: float = 1.0,
                 **kwagrs):
        super().__init__(**kwagrs)
        self.input_dim = input_dim
        self.num_formulas = num_formulas
        self.keep_feature_prob_arr = keep_feature_prob_arr
        self.elastic_net_beta = elastic_net_beta
        self.binary_threshold_eps = binary_threshold_eps
        self.v_temp = tf.Variable(tf.zeros(self.num_formulas), trainable=False)
        self.elastic_net_alpha = None
        self.learnable_mask = None

    def build(self, input_shape):
        self.elastic_net_alpha = tf.Variable(initial_value=tf.constant(0.),
                                             name='elastic_net_alpha')
        self.learnable_mask = tf.Variable(initial_value=tf.fill(dims=(self.input_dim, self.num_formulas),
                                                                value=self.binary_threshold_eps + 0.5),
                                          shape=(self.input_dim, self.num_formulas),
                                          name='learnable_mask')

    def call(self,
             num_literals_per_formula_arr=None,
             ):
        literals_random_mask, formulas_random_mask = generate_random_mask(self.input_dim,
                                                                          self.keep_feature_prob_arr,
                                                                          num_literals_per_formula_arr,
                                                                          self.num_formulas)
        num_effective_features = tf.reduce_sum(formulas_random_mask, axis=0)
        formulas_random_mask = tf.reshape(formulas_random_mask, shape=tf.shape(self.learnable_mask))
        ext_matrix = extension_matrix(self.v_temp,
                                      self.num_formulas,
                                      num_literals_per_formula_arr)
        learnable_mask_01 = binary_threshold(self.learnable_mask, eps=self.binary_threshold_eps)

        l2_square_norm_selected = tf.linalg.diag_part(tf.matmul(tf.transpose(tf.square(self.learnable_mask)),
                                                                formulas_random_mask))
        l1_norm_selected = tf.linalg.diag_part(tf.matmul(tf.transpose(tf.abs(self.learnable_mask)),
                                                         formulas_random_mask))

        l2 = tf.abs(
            tf.divide(l2_square_norm_selected, num_effective_features) - self.elastic_net_beta * self.binary_threshold_eps ** 2)
        l1 = tf.abs(tf.divide(l1_norm_selected, num_effective_features) - self.elastic_net_beta * self.binary_threshold_eps)
        elastic_net_reg = tf.reduce_mean(
            (l2 * ((1 - tf.nn.sigmoid(self.elastic_net_alpha)) / 2) + l1 * tf.nn.sigmoid(self.elastic_net_alpha)))
        learnable_binary_mask = tf_matmul_sparse_dense(ext_matrix, learnable_mask_01)
        return learnable_binary_mask, literals_random_mask, elastic_net_reg

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'num_formulas': self.num_formulas,
                       'keep_feature_prob_arr': self.keep_feature_prob_arr,
                       'elastic_net_beta': self.elastic_net_beta,
                       'binary_threshold_eps': self.binary_threshold_eps,
                       'v_temp': self.v_temp})
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim, **config)



from tensorflow import keras
from teras.layerflow.layers.saint.saint_transformer import SAINTTransformer as _SAINTTransformerLF
from teras.layers.saint.multi_head_inter_sample_attention import MultiHeadInterSampleAttention
from teras.layers.common.transformer import Transformer, FeedForward


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class SAINTTransformer(_SAINTTransformerLF):
    """
    SAINT Transformer layer as proposed by Gowthami Somepalli et al.
    in the paper SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.
    It differs from the usual Transformer (L) block in that it contains additional
    multihead intersample attention layer in addition to the usual multihead attention layer

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        data_dim: ``int``,
            Dimensionality of the input dataset,
            or the total number of features in the dataset.

        embedding_dim: ``int``, default 32,
            Embedding dimensions used to embedd numerical and
            categorical features. These server as the key dimensions
            in the MultiHeadAttention layer.

        num_attention_heads: ``int``, default 8,
            Number of heads to use in the typical ``MultiHeadAttention``
            layer that will be applied over features.

        num_inter_sample_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadInterSampleAttention``
            that will be applied over rows

        embedding_dim: ``int``, default 32,
            Embedding dimensions. These will also serve as key dimensions
             for the attention layers

        attention_dropout: ``float``, default 0.1,
            Dropout rate for ``MultiHeadAttention`` which is applied over
             features.

        inter_sample_attention_dropout: ``float``, default 0.1,
            Dropout rate for ``MultiHeadInterSampleAttention`` which is
            applied over rows.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate for the ``Dropout`` layer that is part of the
            ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the ``embedding_dim``
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the ``FeedForward`` layer.

        apply_attention_to_features: ``bool``, default True,
            Whether to apply attention over features.
            If True, the regular ``MultiHeadAttention`` layer will be applied
            over features.

        apply_attention_to_rows: ``bool``, default True,
            Whether to apply attention over rows.
            If True, the ``MultiHeadInterSampleAttention`` will apply attention
            over rows.
            NOTE: It is strongly recommended to keep both as True, but you
            can turn one off for experiment's sake.
            Also, note that, both CANNOT be False at the same time!
    """
    def __init__(self,
                 data_dim: int,
                 embedding_dim: int = 32,
                 num_attention_heads: int = 8,
                 num_inter_sample_attention_heads: int = 8,
                 attention_dropout: float = 0.1,
                 inter_sample_attention_dropout: float = 0.1,
                 feedforward_dropout: float = 0.1,
                 feedforward_multiplier: int = 4,
                 norm_epsilon: float = 1e-6,
                 apply_attention_to_features: bool = True,
                 apply_attention_to_rows: bool = True,
                 **kwargs):
        multihead_inter_sample_attention = MultiHeadInterSampleAttention(
            num_heads=num_inter_sample_attention_heads,
            key_dim=embedding_dim * data_dim,
            dropout=inter_sample_attention_dropout,
            name="inter_sample_multihead_attention"
        )
        feed_forward = FeedForward(embedding_dim=embedding_dim,
                                   multiplier=feedforward_multiplier,
                                   dropout=feedforward_dropout)
        transformer = Transformer(embedding_dim=embedding_dim,
                                  num_attention_heads=num_attention_heads,
                                  attention_dropout=attention_dropout,
                                  feedforward_dropout=feedforward_dropout,
                                  feedforward_multiplier=feedforward_multiplier,
                                  norm_epsilon=norm_epsilon,
                                  name="inner_trasnformer_block_for_features")

        super().__init__(multi_head_inter_sample_attention=multihead_inter_sample_attention,
                         feed_forward=feed_forward,
                         transformer=transformer,
                         **kwargs)

        self.data_dim = data_dim
        self.embedding_dim = embedding_dim
        self.num_attention_heads = num_attention_heads
        self.num_inter_sample_attention_heads = num_inter_sample_attention_heads
        self.attention_dropout = attention_dropout
        self.inter_sample_attention_dropout = inter_sample_attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.norm_epsilon = norm_epsilon
        self.apply_attention_to_features = apply_attention_to_features
        self.apply_attention_to_rows = apply_attention_to_rows

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'data_dim': self.data_dim,
                  'embedding_dim': self.embedding_dim,
                  'num_attention_heads': self.num_attention_heads,
                  'num_inter_sample_attention_heads': self.num_inter_sample_attention_heads,
                  'attention_dropout': self.attention_dropout,
                  'inter_sample_attention_dropout': self.inter_sample_attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'norm_epsilon': self.norm_epsilon,
                  'apply_attention_to_features': self.apply_attention_to_features,
                  'apply_attention_to_rows': self.apply_attention_to_rows,
                  'num_embedded_features': self.num_embedded_features,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        # data_dim is the only positional argument
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim, **config)



import tensorflow as tf
from tensorflow import keras
from teras.utils.pcgain import cluster
import tensorflow_probability as tfp
from teras.losses.pcgain import (generator_loss,
                                 generator_pretraining_loss,
                                 discriminator_loss)


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class PCGAIN(keras.Model):
    """
    PCGAIN model with LayerFlow design.
    PCGAIN is a missing data imputation model based
    on the GAIN architecture.

    This implementation is based on the architecture
    proposed by Yufeng Wang et al. in the paper
    "PC-GAIN: Pseudo-label Conditional Generative Adversarial
    Imputation Networks for Incomplete Data"

    Reference(s):
        https://arxiv.org/abs/2011.07770

    Args:
        pretrainer: ``keras.Model``,
            An instance of ``GAIN`` model or any custom keras model that can be used
            as a pretrainer to pretrain the ``PCGAINGenerator`` and ``PCGAINDiscriminator`` instances.
            The pretrainer is used to pretrain the Generator and Discriminators, and to impute
            the pretraining dataset, which is then clustered to generated pseudo labels,
            which combined with the imputed data are used to train the classifier.

            IMPORTANT NOTE:
                If the ``PCGAINGenerator`` and ``PCGAINDiscriminator`` instances being used by
                the ``pretrainer`` will be used by the ``PCGAIN`` model.
                Hence, if you're using a custom Pretrainer model, it must have attributes or
                proporties named ``generator`` and ``discriminator`` to fetch the Generator and
                Discriminator instances repsectively.

        classifier: ``keras.Model``,
            An instance of the ``PCGAINClassifier`` model or any custom model that can work
            in its palce.
            It is used in the training of the Generator component
            of the PCGAIN architecture after it has been pretrained by the ``pretrainer``.
            The ``classifier`` itself is trained on the imputed data in the pretraining step
            combined with the pseudo labels generated for the imputed data by clustering.
            You can import the ``PCGAINClassifier`` as follows,
                >>> from teras.models import PCGAINClassifier

        num_discriminator_steps: ``int``, default 1,
            Number of discriminator training steps per PCGAIN training step.

        hint_rate: ``float``,
            Hint rate will be used to sample binary vectors for
            ``hint vectors`` generation. Must be between 0. and 1.
            Hint vectors ensure that generated samples follow the
            underlying data distribution.

        alpha: ``float``, default 200.,
            Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely, ``generator_loss = cross_entropy_loss + alpha * mse_loss``
            The higher the ``alpha``, the more the mse_loss will affect the
            overall generator loss.

        beta: ``float``, default 100.,
            Hyper parameter for generator loss computation that
            controls the contribution of the classifier's loss to the
            overall generator loss.

        num_clusters: ``int``, default 5,
            Number of clusters to cluster the imputed data
            that is generated during pretraining.
            These clusters serve as pseudo labels for training of classifier.

        clustering_method: ``str``, default "kmeans",
            Should be one of the following,
            ["Agglomerative", "KMeans", "MiniBatchKMeans", "Spectral", "SpectralBiclustering"]
            The names are case in-sensitive.
    """
    def __init__(self,
                 pretrainer: keras.Model = None,
                 classifier: keras.Model = None,
                 num_discriminator_steps: int = 1,
                 hint_rate: float = 0.9,
                 alpha: float = 200,
                 beta: float = 100,
                 num_clusters: int = 5,
                 clustering_method: str = "kmeans",
                 **kwargs):
        super().__init__(**kwargs)
        if not hasattr(pretrainer, "generator"):
            raise ValueError("The `pretrainer` doesn't have a `generator` attribute or property. \n"
                             "`pretrainer` must have an `generator` attribute or property "
                             "to fetch the `Generator` instance.")

        if not hasattr(pretrainer, "discriminator"):
            raise ValueError("The `pretrainer` doesn't have a `discriminator` attribute or property. \n"
                             "`pretrainer` must have an `discriminator` attribute or property "
                             "to fetch the `Discriminator` instance.")

        if not 0. <= hint_rate <= 1.0:
            raise ValueError("`hint_rate` should be between 0. and 1. "
                             f"Received {hint_rate}.")

        # If user has passed a pretrainer it must have generator and discriminator instances as attributes
        # which it will pretrian.
        # We use the same generator and discriminator instances from the pretrainer.
        self.pretrainer = pretrainer
        self.classifier = classifier
        self.num_discriminator_steps = num_discriminator_steps
        self.hint_rate = hint_rate
        self.alpha = alpha
        self.beta = beta
        self.num_clusters = num_clusters
        self.clustering_method = clustering_method

        self.generator = self.pretrainer.generator
        self.discriminator = self.pretrainer.discriminator

        self.z_sampler = tfp.distributions.Uniform(low=0.,
                                                   high=0.01,
                                                   name="z_sampler")
        self.hint_vectors_sampler = tfp.distributions.Binomial(total_count=1,
                                                               probs=self.hint_rate,
                                                               name="hint_vectors_generator")

        # Flag to keep track of whether generator and discriminator
        # have been pretrained or not
        self._pretrained = False
        # Flag to check for training start and call specific functions
        self._first_batch = True

        # Loss trackers
        self.generator_loss_tracker = keras.metrics.Mean(name="generator_loss")
        self.discriminator_loss_tracker = keras.metrics.Mean(name="discriminator_loss")

    def compile(self,
                generator_optimizer=keras.optimizers.Adam(),
                discriminator_optimizer=keras.optimizers.Adam(),
                generator_loss=generator_loss,
                discriminator_loss=discriminator_loss,
                generator_pretraining_loss=generator_pretraining_loss,
                classifier_loss=keras.losses.SparseCategoricalCrossentropy(),
                classifier_optimizer=keras.optimizers.RMSprop(),
                ):
        super().compile()
        self.generator_optimizer = generator_optimizer
        self.discriminator_optimizer = discriminator_optimizer
        self.generator_loss = generator_loss
        self.discriminator_loss = discriminator_loss
        self.generator_pretraining_loss = generator_pretraining_loss
        # compile the pretrainer (which is GAIN model)
        # using these losses.
        # Note that only generator has a different pretraining loss,
        # as the discriminator uses the same loss!
        self.pretrainer.compile(generator_optimizer=generator_optimizer,
                                discriminator_optimizer=discriminator_optimizer,
                                generator_loss=generator_pretraining_loss,
                                discriminator_loss=discriminator_loss
                                )

        self.classifier.compile(loss=classifier_loss,
                                optimizer=classifier_optimizer)

    def get_generator(self):
        return self.generator

    def get_discriminator(self):
        return self.discriminator

    def call(self, inputs, mask=None, training=None):
        if mask is not None:
            inputs = tf.concat([inputs, mask], axis=1)
        return self.generator(inputs)

    def pretrain(self,
                 pretraining_dataset=None,
                 pretrainer_fit_kwargs: dict = {},
                 classifier_fit_kwargs: dict = {}):
        """
        Pretrain function that does everything required before PC-GAIN training.
        It does the following in the given order:
            1. Pretrains the Generator and Discriminator and imputes pretraining dataset.
            2. Using clustering on imputed data, generates pseudo labels.
            3. Trains the classifier.

        Terminology Note:
            `Pretrainer` refers to the model used to pretrain Discriminator and Generator.
            `Classifier` refers to the model trained on imputed data and pseudo labels.

        Args:
            pretraining_dataset: Get it from DataSampler's `get_pretraining_dataset` method.
            pretrainer_fit_kwargs: Dictionary of keyword arguments for pretrainer's fit method.
            classifier_fit_kwargs: Dictionary of keyword arguments for classifier's fit method.

        p.s. I know accepting arguments as dict is not a good design but for the time being,
        please tolerate it. I'll fix it in the near future!
        """
        print("Pretrain Step 1/4: Pretraining the generator and discriminator")
        self.pretrainer.fit(pretraining_dataset, **pretrainer_fit_kwargs)

        print("\nPretrain Step 2/4: Imputing data")
        imputed_data = self.pretrainer.predict(pretraining_dataset)
        self.generator = self.pretrainer.generator
        self.discriminator = self.pretrainer.discriminator

        print("\nPretrain Step 3/4: Generating pseudo labels")
        pseudo_labels = cluster(x=imputed_data,
                                num_clusters=self.num_clusters,
                                method=self.clustering_method)

        print("\nPretrain Step 4/4: Training the classifier")
        self.classifier.fit(imputed_data, y=pseudo_labels, **classifier_fit_kwargs)

        print("\nPretraining completed successfully!")
        self._pretrained = True

        print("\nTraining PCGAIN")

    # We'll utilize the same DataTransformer and DataSampler classes from
    # the GAIN architecture. and hence the incoming data follows the same format.
    def train_step(self, data):
        if self._first_batch:
            if not self._pretrained:
                raise AssertionError("The model has not yet been pretrained. "
                                     "PC-GAIN requires Generator and Discriminator "
                                     "to be pretrained before the main training. "
                                     "You must call the `pretrain` method before "
                                     "calling the `fit` method.")
            self._first_batch = False

        # data is a tuple of x_generator and x_discriminator batches
        # drawn from the dataset. The reason behind generating two separate
        # batches of data at each step is that PCGAIN requires separate
        # batches of data for Discriminator and Generator.
        x_gen, x_disc = data

        # ====> Training the Discriminator <=====
        for _ in range(self.num_discriminator_steps):
            # Create mask
            mask = tf.constant(1.) - tf.cast(tf.math.is_nan(x_disc), dtype=tf.float32)
            # replace nans with 0.
            x_disc = tf.where(tf.math.is_nan(x_disc), x=0., y=x_disc)
            # Sample noise
            z = self.z_sampler.sample(sample_shape=tf.shape(x_disc))
            # Sample hint vectors
            hint_vectors = self.hint_vectors_sampler.sample(sample_shape=(tf.shape(x_disc)))
            hint_vectors *= mask
            # Combine random vectors with original data
            x_disc = x_disc * mask + (1 - mask) * z
            # Generate samples
            generated_samples = self(x_disc, mask=mask)
            # Combine generated samples with original data
            x_hat_disc = (generated_samples * (1 - mask)) + (x_disc * mask)
            with tf.GradientTape() as tape:
                discriminator_pred = self.discriminator(tf.concat([x_hat_disc, hint_vectors], axis=1))
                loss_disc = self.discriminator_loss(discriminator_pred, mask)
            gradients = tape.gradient(loss_disc, self.discriminator.trainable_weights)
            self.discriminator_optimizer.apply_gradients(zip(gradients, self.discriminator.trainable_weights))

        # =====> Train the Generator <=====
        mask = tf.constant(1.) - tf.cast(tf.math.is_nan(x_gen), dtype=tf.float32)
        x_gen = tf.where(tf.math.is_nan(x_gen), x=0., y=x_gen)
        z = self.z_sampler.sample(sample_shape=tf.shape(x_gen))
        hint_vectors = self.hint_vectors_sampler.sample(sample_shape=(tf.shape(x_gen)))
        hint_vectors *= mask
        x_gen = x_gen * mask + (1 - mask) * z
        # "Introducing Pseudo label supervision"
        # we don't need to use one hot, since we're using
        # softmax activation at the end and it gives us
        # probability for each class
        with tf.GradientTape() as tape:
            generated_samples = self(x_gen, mask=mask)
            # Combine generated samples with original/observed data
            x_hat = (generated_samples * (1 - mask)) + (x_gen * mask)
            classifier_preds = self.classifier(x_hat)
            disc_preds = self.discriminator(tf.concat([x_hat, hint_vectors], axis=1))
            loss_gen = self.generator_loss(generated_samples=generated_samples,
                                           real_samples=x_gen,
                                           discriminator_pred=disc_preds,
                                           mask=mask,
                                           alpha=self.alpha,
                                           beta=self.beta,
                                           classifier_pred=classifier_preds)
        gradients = tape.gradient(loss_gen, self.generator.trainable_weights)
        self.generator_optimizer.apply_gradients(zip(gradients, self.generator.trainable_weights))

        # Update custom tracking metrics
        self.generator_loss_tracker.update_state(loss_gen)
        self.discriminator_loss_tracker.update_state(loss_disc)

        results = {m.name: m.result() for m in self.metrics}
        results.update({"generator_" + m.name: m.result() for m in self.generator.metrics})
        results.update({"discriminator_" + m.name: m.result() for m in self.discriminator.metrics})
        return results

    def predict_step(self, data):
        """
        Args:
            Transformed data.
        Returns:
            Imputed data that should be reverse transformed
            to its original form.
        """
        if isinstance(data, tuple):
            data = data[0]
        data = tf.cast(data, tf.float32)
        # Create mask
        mask = tf.constant(1.) - tf.cast(tf.math.is_nan(data), dtype=tf.float32)
        data = tf.where(tf.math.is_nan(data), x=0., y=data)
        # Sample noise
        z = self.z_sampler.sample(sample_shape=tf.shape(data))
        x = mask * data + (1 - mask) * z
        imputed_data = self(x, mask=mask)
        imputed_data = mask * data + (1 - mask) * imputed_data
        return imputed_data

    def impute(self,
               x,
               data_transformer=None,
               reverse_transform=True,
               batch_size=None,
               verbose="auto",
               steps=None,
               callbacks=None,
               max_queue_size=10,
               workers=1,
               use_multiprocessing=False
               ):
        """
        Impute function combines PC-GAIN's `predict` method and
        DataTransformer's `reverse_transform` method to fill
        the missing data and transform into the original format.
        It exposes all the arguments taken by the `predict` method.

        Args:
            x: ``pd.DataFrame``, dataset with missing values.

            data_transformer:
                An instance of ``PCGAINDataTransformer`` class which
                was used in transforming the raw input data

            reverse_transform: ``bool``, default True,
                whether to reverse
                transformed the raw imputed data to original format.

        Returns:
            Imputed data in the original format.
        """
        x_imputed = self.predict(x,
                                 batch_size=batch_size,
                                 verbose=verbose,
                                 steps=steps,
                                 callbacks=callbacks,
                                 max_queue_size=max_queue_size,
                                 workers=workers,
                                 use_multiprocessing=use_multiprocessing)
        if reverse_transform:
            if data_transformer is None:
                raise ValueError("To reverse transform the raw imputed data, `data_transformer` must not be None. "
                                 "Please pass the instance of DataTransformer class used to transform the input "
                                 "data as argument to this `impute` method."
                                 "Or alternatively, you can set `reverse_transform` parameter to False, "
                                 "and manually reverse transform the generated raw data to original format "
                                 "using the `reverse_transform` method of DataTransformer instance.")
            x_imputed = data_transformer.reverse_transform(x_imputed)
        return x_imputed

    def get_config(self):
        config = super().get_config()
        config.update({"pretrainer": keras.layers.serialize(self.pretrainer),
                       "classifier": keras.layers.serialize(self.classifier),
                       "num_discriminator_steps": self.num_discriminator_steps,
                       "hint_rate": self.hint_rate,
                       "alpha": self.alpha,
                       "beta": self.beta,
                       "num_clusters": self.num_clusters,
                       "clustering_method": self.clustering_method,
                       }
                      )
        return config



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class SAINTNumericalFeatureEmbedding(keras.layers.Layer):
    """
    SAINTNumericalFeatureEmbedding layer based on the architecture proposed
    by Gowthami Somepalli et al. in the paper
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      numerical_features,
                ..                                                      categorical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimension is the dimensionality of the output layer or
            the dimensionality of the embeddings produced.
            (These embedding dimensions are the same used for the embedding categorical features)

        hidden_dim: ``int``, default 16,
            Hidden dimension, used by the first dense layer i.e the hidden layer whose outputs
            are later projected to the ``emebedding_dim``
    """
    def __init__(self,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 hidden_dim: int = 16,
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim

        self._num_numerical_features = len(self.features_metadata["numerical"])
        # Need to create as many embedding layers as there are numerical features
        self.embedding_layers = []
        for _ in range(self._num_numerical_features):
            self.embedding_layers.append(
                keras.models.Sequential([
                    keras.layers.Dense(units=self.hidden_dim, activation="relu"),
                    keras.layers.Dense(units=self.embedding_dim)
                ]
                )
            )

    def call(self, inputs):
        numerical_feature_embeddings = tf.TensorArray(size=self._num_numerical_features,
                                                      dtype=tf.float32)

        for i, (feature_name, feature_idx) in enumerate(self.features_metadata["numerical"].items()):
            feature = tf.expand_dims(inputs[:, feature_idx], 1)
            embedding = self.embedding_layers[i]
            feature = embedding(feature)
            numerical_feature_embeddings = numerical_feature_embeddings.write(i, feature)

        numerical_feature_embeddings = tf.squeeze(numerical_feature_embeddings.stack())
        numerical_feature_embeddings = tf.transpose(numerical_feature_embeddings, perm=[1, 0, 2])
        numerical_feature_embeddings.set_shape((None, self._num_numerical_features, self.embedding_dim))
        return numerical_feature_embeddings

    def get_config(self):
        config = super().get_config()
        config.update({'features_metadata': self.features_metadata,
                       'embedding_dim': self.embedding_dim,
                       'hidden_dim': self.hidden_dim,
                       })
        return config



import tensorflow as tf
from teras.layers.dnfnet.dnfnet_feature_selection import DNFNetFeatureSelection
from teras.utils.dnfnet import compute_num_literals_per_formula


def test_dnfnet_feature_selection_valid_call():
    feature_selection = DNFNetFeatureSelection(input_dim=10,
                                               num_formulas=16)
    inputs = tf.ones((8, 10), dtype=tf.float32)
    num_conjunctions_arr = [6, 9, 12, 15]
    conjunctions_depth_arr = [2, 4, 6]
    num_literals_per_formula_arr = compute_num_literals_per_formula(num_conjunctions_arr,
                                                                    conjunctions_depth_arr)

    learnable_binary_mask, literals_random_mask, elastic_net_reg = feature_selection(num_literals_per_formula_arr)


def test_dnfnet_feature_selection_output_shape():
    feature_selection = DNFNetFeatureSelection(input_dim=10,
                                               num_formulas=16)
    inputs = tf.ones((8, 10), dtype=tf.float32)
    num_conjunctions_arr = [6, 9, 12, 15]
    conjunctions_depth_arr = [2, 4, 6]
    num_literals_per_formula_arr = compute_num_literals_per_formula(num_conjunctions_arr,
                                                                    conjunctions_depth_arr)

    learnable_binary_mask, literals_random_mask, elastic_net_reg = feature_selection(num_literals_per_formula_arr)
    assert len(tf.shape(learnable_binary_mask)) == 2
    assert tf.shape(learnable_binary_mask)[0] == 10
    assert tf.shape(learnable_binary_mask)[1] == 0
    assert len(tf.shape(literals_random_mask)) == 2
    assert tf.shape(literals_random_mask)[0] == 10
    assert tf.shape(literals_random_mask)[1] == 0
    assert len(tf.shape(elastic_net_reg)) == 0



import tensorflow as tf
from tensorflow import keras
from teras.layerflow.models.saint import (SAINT,
                                          SAINTPretrainer)
from teras.layers import (CategoricalFeatureEmbedding,
                          SAINTNumericalFeatureEmbedding,
                          SAINTEncoder,
                          MixUp,
                          CutMix,
                          SAINTReconstructionHead,
                          SAINTProjectionHead)
from teras.utils import get_features_metadata_for_embedding
import pandas as pd
import numpy as np
import os


class SAINTTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.input_dim = 4
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                         embedding_dim=32)
        self.numerical_feature_embedding = SAINTNumericalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                          embedding_dim=32)
        self.encoder = SAINTEncoder(data_dim=self.input_dim)
        self.head = keras.layers.Dense(1)
        self.data_batch = data.values

    def test_valid_call(self):
        model = SAINT(input_dim=self.input_dim,
                      categorical_feature_embedding=self.categorical_feature_embedding,
                      numerical_feature_embedding=self.numerical_feature_embedding,
                      encoder=self.encoder,
                      head=self.head)
        model(self.data_batch)

    def test_save_and_load(self):
        model = SAINT(input_dim=self.input_dim,
                      categorical_feature_embedding=self.categorical_feature_embedding,
                      numerical_feature_embedding=self.numerical_feature_embedding,
                      encoder=self.encoder,
                      head=self.head)
        save_path = os.path.join(self.get_temp_dir(), "saint_lf.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)
        self.assertAllClose(outputs_original, outputs_reloaded)


class SAINTPretrainerTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.input_dim = 4
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                         embedding_dim=32)
        self.numerical_feature_embedding = SAINTNumericalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                          embedding_dim=32)
        self.encoder = SAINTEncoder(data_dim=self.input_dim)

        # layers for pretrainer
        self.mixup = MixUp()
        self.cutmix = CutMix()
        self.projection_head_1 = SAINTProjectionHead()
        self.projection_head_2 = SAINTProjectionHead()
        self.reconstruction_head = SAINTReconstructionHead(features_metadata=self.features_metadata)
        self.data_batch = data.values

    def test_valid_call(self):
        base_model = SAINT(input_dim=self.input_dim,
                           categorical_feature_embedding=self.categorical_feature_embedding,
                           numerical_feature_embedding=self.numerical_feature_embedding,
                           encoder=self.encoder)
        pretrainer = SAINTPretrainer(model=base_model,
                                     features_metadata=self.features_metadata,
                                     mixup=self.mixup,
                                     cutmix=self.cutmix,
                                     projection_head_1=self.projection_head_1,
                                     projection_head_2=self.projection_head_2,
                                     reconstruction_head=self.reconstruction_head)
        pretrainer(self.data_batch)

    def test_save_and_load(self):
        base_model = SAINT(input_dim=self.input_dim,
                           categorical_feature_embedding=self.categorical_feature_embedding,
                           numerical_feature_embedding=self.numerical_feature_embedding,
                           encoder=self.encoder)
        pretrainer = SAINTPretrainer(model=base_model,
                                     features_metadata=self.features_metadata,
                                     mixup=self.mixup,
                                     cutmix=self.cutmix,
                                     projection_head_1=self.projection_head_1,
                                     projection_head_2=self.projection_head_2,
                                     reconstruction_head=self.reconstruction_head)
        save_path = os.path.join(self.get_temp_dir(), "saint_pretrainer_lf.keras")
        pretrainer.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = pretrainer(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)
        # We can't check for AllClose



from tensorflow import keras
from teras.utils.types import LayersCollection


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class DNFNet(keras.Model):
    """
    DNFNet model based on the DNFNet architecture with LayerFlow design.
    DNFNet architecture is proposed by Liran Katzir et al.
    in the paper,
    "NET-DNF: Effective Deep Modeling Of Tabular Data."

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        dnnf_layers: ``List[layers.Layer]`` or ``keras.Model``,
            A list of instances of ``DNNF`` layers or
            a Keras layer or model made up of ``DNNF`` layers.
            You can import the ``DNNF`` layer as follows,
                >>> from teras.layers import DNNF

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.
            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """
    def __init__(self,
                 input_dim: int,
                 dnnf_layers: LayersCollection,
                 head: keras.layers.Layer = None,
                 **kwargs):
        if isinstance(dnnf_layers, (keras.layers.Layer, keras.models.Model)):
            # keep it as is
            dnnf_layers = dnnf_layers
        elif isinstance(dnnf_layers, (list, tuple)):
            dnnf_layers = keras.models.Sequential(dnnf_layers,
                                                  name="dnnf_layers")
        else:
            raise ValueError("`dnnf_layers` can either be a list of `DNNF` layers, a Keras Layer, "
                             f"or a Keras model model but received type: {type(dnnf_layers)}.")
        inputs = keras.layers.Input(shape=(input_dim,))
        outputs = dnnf_layers(inputs)
        if head is not None:
            outputs = head(outputs)
        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)
        self.input_dim = input_dim
        self.dnnf_layers = dnnf_layers
        self.head = head

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'dnnf_layers': keras.layers.serialize(self.dnnf_layers),
                       'head': keras.layers.serialize(self.head)
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        dnnf_layers = keras.layers.deserialize(config.pop("dnnf_layers"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   dnnf_layers=dnnf_layers,
                   head=head,
                   **config)



import tensorflow as tf
from tensorflow import keras
from sklearn.model_selection import KFold, StratifiedKFold
import numpy as np
from typing import List

MODELS_LIST = List[keras.Model]

# NOTE: This is a very simple implementation of Bagging.
# And mostly is for demonstration purpose.
# Will make it efficient and better soon enough.
# Also, currently there's just Bagging class, but my plan is to add
# BaggingClassifier and BaggingRegressor like sklearn.


class Stacking:
    def __int__(self,
                base_learners: MODELS_LIST,
                num_splits: int = 5
                ):
        self.base_learners = base_learners
        self.num_splits = num_splits

    def fit(self, X, y):
        self.meta_data = np.zeros((len(self.base_learners), len(X)))
        self.meta_targets = np.zeros(len(X))

        kf = KFold(n_splits=self.num_splits, shuffle=True)
        meta_index = 0

        for train_idx, test_idx in kf.split(X, y):
            X_train, y_train = X[train_idx], y[train_idx]
            X_test, y_test = X[test_idx], y[test_idx]
            for i in range(len(self.base_learners)):
                learner = self.base_learners[i]
                learner.fit(X_train, y_train)
                predictions = learner.predict(X_test)
                self.meta_data[i][meta_index:meta_index + len(test_idx)] = predictions

            self.meta_targets[meta_index:meta_index + len(test_idx)] = y_test
            meta_index += len(test_idx)

        return self.meta_data, self.meta_targets


from tensorflow import keras
from tensorflow.keras import backend as K
from teras.layerflow.layers.common.common import HiLOL
from teras.utils import (serialize_layers_collection,
                         deserialize_layers_collection)
from typing import List

LIST_OF_LAYERS = List[keras.layers.Layer]


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.saint")
class SAINTReconstructionHeadBlock(HiLOL):
    """
    SAINTReconstructionHeadBlock layer with LayerFlow design.
    It is used in constructing SAINTReconstructionHead.
    One ``SAINTReconstructionHeadBlock`` is created for each feature in the dataset.

    Args:
        hidden_block: ``keras.layers.Layer``,
            An instance of ``Dense`` layer, or any custom layer that can
            serve as the hidden block.

        output_layer: ``keras.layers.Layer``,
            Any layer that can serve as the output layer BUT it must have
            an output dimensionality equal to the dimensionality of the feature
            it will be applied to. For categorical features, it is equal to the
            number of classes in the feature, and for numerical features,
            it is equal to 1.
    """
    def __init__(self,
                 hidden_block: keras.layers.Layer,
                 output_layer: keras.layers.Layer,
                 **kwargs):
        super().__init__(hidden_block=hidden_block,
                         output_layer=output_layer,
                         **kwargs)


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.saint")
class SAINTReconstructionHead(keras.layers.Layer):
    """
    SAINTReconstructionHead layer with LayerFlow desing for ``SAINTPretrainer`` model.
    SAINT applies a separate single hidden layer MLP block
    (here we name it, the ReconstructionBlock)
    with an output layer where output dimensions are equal
    to the number of categories in the case of categorical
    features and 1 in the case of numerical features.

    Args:
        reconstruction_blocks: ``List[keras.layers.Layer]``,
            A list of `SAINTRReconstructionHeadBlock` layers - one for each feature,
            where the ``SAINTRReconstructionHeadBlock`` has dimensionality equal to the cardinality
            of that feature.
            For instance, for a categorical feature, the dimensionality of ``SAINTReconstructionBlock``
            will be equal to the number of classes in that feature, while for a numerical feature
            it is just equal to 1.
            You can import the ``SAINTRReconstructionHeadBlock`` layer as follows,
                >>> from teras.layerflow.layers import SAINTReconstructionHeadBlock
    """
    def __init__(self,
                 reconstruction_blocks: LIST_OF_LAYERS,
                 **kwargs):
        super().__init__(**kwargs)
        self.reconstruction_blocks = reconstruction_blocks

    def call(self, inputs):
        """
        Args:
            inputs: SAINT transformer outputs for the augmented data.
                Since we apply categorical and numerical embedding layers
                separately and then combine them into a new features matrix
                this effectively makes the first k features in the outputs
                categorical (since categorical embeddings are applied first)
                and all other features numerical.
                Here, k = num_categorical_features

        Returns:
            Reconstructed input features
        """
        reconstructed_inputs = []
        for idx, block in enumerate(self.reconstruction_blocks):
            feature_encoding = inputs[:, idx]
            reconstructed_feature = block(feature_encoding)
            reconstructed_inputs.append(reconstructed_feature)
        # the reconstructed inputs will have features equal to
        # `number of numerical features` + `number of categories in the categorical features`
        reconstructed_inputs = K.concatenate(reconstructed_inputs, axis=1)
        return reconstructed_inputs

    def get_config(self):
        config = super().get_config()
        config.update({'reconstruction_blocks': serialize_layers_collection(self.reconstruction_blocks),
                       })
        return config

    @classmethod
    def from_config(cls, config):
        reconstruction_blocks = deserialize_layers_collection(config.pop("reconstruction_blocks"))
        return cls(reconstruction_blocks, **config)



from sklearn.cluster import (AgglomerativeClustering,
                             KMeans,
                             MiniBatchKMeans,
                             SpectralClustering,
                             SpectralBiclustering)


SUPPORTED_CLUSTERING_METHODS = ["Agglomerative", "KMeans", "MiniBatchKMeans", "Spectral", "SpectralBiclustering"]


def cluster(x, num_clusters=5, method="kmeans"):
    """
    Clusters the data into `num_clusters` clusters using
    the specified cluster method.

    Args:
        x: Dataset to cluster
        num_clusters: Number of clusters to cluster the data into.
        method: Should be one of the following,
            ["Agglomerative", "KMeans", "MiniBatchKMeans", "Spectral", "SpectralBiclustering"]
            The names are case in-sensitive.
            Defaults to "kmeans".

    Returns:
        Predicted cluster index for each sample.
    """
    method = method.lower()
    if method == "agglomerative":
        cluster_idx = AgglomerativeClustering(n_clusters=num_clusters).fit_predict(x)
    elif method == "kmeans":
        cluster_idx = KMeans(n_clusters=num_clusters).fit_predict(x)
    elif method == "minibatchkmeans":
        cluster_idx = MiniBatchKMeans(n_clusters=num_clusters).fit_predict(x)
    elif method == "spectral":
        cluster_idx = SpectralClustering(n_clusters=num_clusters).fit_predict(x)
    elif method == "spectralbiclustering":
        cluster_idx = SpectralBiclustering(n_clusters=num_clusters).fit(x).row_labels_
    else:
        raise ValueError(f"The clustering method name passed `{method}` is either not "
                         f"currently supported or you made a typo in the name."
                         f"Please note that ")
    return cluster_idx





from tensorflow import keras
from teras.layers.rtdl_resnet.rtdl_resnet_block import RTDLResNetBlock
from teras.layerflow.models.rtdl_resnet import RTDLResNet as _RTDLResNetLF
from teras.utils.types import (ActivationType,
                               NormalizationType,
                               UnitsValuesType)
from teras.layers.common.head import ClassificationHead, RegressionHead


@keras.saving.register_keras_serializable(package="teras.models")
class RTDLResNet(_RTDLResNetLF):
    """
    The ResNet model based on the architecture proposed by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        input_dim: ``int``,
            The dimensionality of the input dataset,
            or the number of features in the input dataset.

        num_blocks: ``int``, default 4,
            Number of ResNet blocks to use.

        block_hidden_dim: ``int``, default 64,
            Dimensionality of the hidden layer in the ResNet block.

        block_dropout_hidden: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the hidden dense layer in the ``RTDLResNetBlock`` layer.

        block_dropout_out: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the output dense layer in the ``RTDLResNetBlock`` layer.

        block_activation_hidden: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the hidden layer in the ``RTDLResNetBlock`` layer.

        block_activation_out: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the output layer in the ``RTDLResNetBlock`` layer.

        block_normalization: ``str`` or ``keras.layers.Layer``, default "batch",
            Normalization layer to normalize the inputs to the ``RTDLResNetBlock`` layer.

        use_skip_connection: ``bool``, default True,
            Whether to use the skip connection in the ``RTDLResNetBlock`` layer.
    """
    def __init__(self,
                 input_dim: int,
                 num_blocks: int = 4,
                 block_hidden_dim: int = 64,
                 block_dropout_hidden: float = 0.,
                 block_dropout_out: float = 0.,
                 block_activation_hidden: ActivationType = "relu",
                 block_activation_out: ActivationType = "relu",
                 block_normalization: NormalizationType = "BatchNormalization",
                 use_skip_connection: bool = True,
                 **kwargs):
        resnet_blocks = keras.models.Sequential(
            [
                RTDLResNetBlock(
                    units=block_hidden_dim,
                    dropout_hidden=block_dropout_hidden,
                    dropout_out=block_dropout_out,
                    activation_hidden=block_activation_hidden,
                    activation_out=block_activation_out,
                    normalization=block_normalization,
                    use_skip_connection=use_skip_connection
                )
                for _ in range(num_blocks)
            ],
            name="resnet_blocks"
        )
        super().__init__(input_dim=input_dim,
                         resnet_blocks=resnet_blocks,
                         **kwargs)
        self.input_dim = input_dim
        self.num_blocks = num_blocks
        self.block_hidden_dim = block_hidden_dim
        self.block_dropout_hidden = block_dropout_hidden
        self.block_dropout_out = block_dropout_out
        self.block_activation_hidden = block_activation_hidden
        self.block_activation_out = block_activation_out
        self.block_normalization = block_normalization
        self.use_skip_connection = use_skip_connection

    def get_config(self):
        activation_hidden_serialized = self.block_activation_hidden
        if not isinstance(self.block_activation_hidden, str):
            activation_hidden_serialized = keras.layers.serialize(self.block_activation_hidden)

        activation_out_serialized = self.block_activation_out
        if not isinstance(self.block_activation_out, str):
            activation_out_serialized = keras.layers.serialize(self.block_activation_out)

        normalization_serialized = self.block_normalization
        if not isinstance(self.block_normalization, str):
            normalization_serialized = keras.layers.serialize(self.block_normalization)

        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'num_blocks': self.num_blocks,
                  'block_hidden_dim': self.block_hidden_dim,
                  'block_dropout_hidden': self.block_dropout_hidden,
                  'block_dropout_out': self.block_dropout_out,
                  'block_activation_hidden': activation_hidden_serialized,
                  'block_activation_out': activation_out_serialized,
                  'block_normalization': normalization_serialized,
                  'use_skip_connection': self.use_skip_connection,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim,
                   **config)


@keras.saving.register_keras_serializable(package="teras.models")
class RTDLResNetClassifier(RTDLResNet):
    """
    The ResNet Classifier model based on the ResNet architecture
    proposed by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict

        head_units_values: ``List[int]`` or ``Tuple[int]``, default [64, 32],
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.

        input_dim: ``int``,
            The dimensionality of the input dataset,
            or the number of features in the input dataset.

        num_blocks: ``int``, default 4,
            Number of ResNet blocks to use.

        block_hidden_dim: ``int``, default 64,
            Dimensionality of the hidden layer in the ResNet block.

        block_dropout_hidden: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the hidden dense layer in the ``RTDLResNetBlock`` layer.

        block_dropout_out: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the output dense layer in the ``RTDLResNetBlock`` layer.

        block_activation_hidden: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the hidden layer in the ``RTDLResNetBlock`` layer.

        block_activation_out: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the output layer in the ``RTDLResNetBlock`` layer.

        block_normalization: ``str`` or ``keras.layers.Layer``, default "batch",
            Normalization layer to normalize the inputs to the ``RTDLResNetBlock`` layer.

        use_skip_connection: ``bool``, default True,
            Whether to use the skip connection in the ``RTDLResNetBlock`` layer.
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 num_blocks: int = 4,
                 block_hidden_dim: int = 64,
                 block_dropout_hidden: float = 0.,
                 block_dropout_out: float = 0.,
                 block_activation_hidden: ActivationType = "relu",
                 block_activation_out: ActivationType = "relu",
                 block_normalization: NormalizationType = "batch",
                 use_skip_connection: bool = True,
                 **kwargs):
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_values,
                                  name="rtdl_resnet_classification_head")
        super().__init__(input_dim=input_dim,
                         num_blocks=num_blocks,
                         block_hidden_dim=block_hidden_dim,
                         block_dropout_hidden=block_dropout_hidden,
                         block_dropout_out=block_dropout_out,
                         block_activation_hidden=block_activation_hidden,
                         block_activation_out=block_activation_out,
                         block_normalization=block_normalization,
                         use_skip_connection=use_skip_connection,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_classes': self.num_classes,
                       'head_units_values': self.head_units_values
                       }
                      )
        return config


@keras.saving.register_keras_serializable(package="teras.models")
class RTDLResNetRegressor(RTDLResNet):
    """
    The ResNet Regressor model based on the ResNet architecture
    proposed by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        num_outputs: ``int``, default 1,
            Number of regression outputs

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            The dimensionality of the input dataset,
            or the number of features in the input dataset.

        num_blocks: ``int``, default 4,
            Number of ResNet blocks to use.

        block_hidden_dim: ``int``, default 64,
            Dimensionality of the hidden layer in the ResNet block.

        block_dropout_hidden: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the hidden dense layer in the ``RTDLResNetBlock`` layer.

        block_dropout_out: ``float``, default 0.,
            Dropout rate to use for the ``Dropout`` layer that is applied
            after the output dense layer in the ``RTDLResNetBlock`` layer.

        block_activation_hidden: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the hidden layer in the ``RTDLResNetBlock`` layer.

        block_activation_out: ``callable`` or ``str`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in the output layer in the ``RTDLResNetBlock`` layer.

        block_normalization: ``str`` or ``keras.layers.Layer``, default "batch",
            Normalization layer to normalize the inputs to the ``RTDLResNetBlock`` layer.

        use_skip_connection: ``bool``, default True,
            Whether to use the skip connection in the ``RTDLResNetBlock`` layer.
    """
    def __init__(self,
                 num_outputs: int = 1,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 num_blocks: int = 4,
                 block_hidden_dim: int = 64,
                 block_dropout_hidden: float = 0.,
                 block_dropout_out: float = 0.,
                 block_activation_hidden: ActivationType = "relu",
                 block_activation_out: ActivationType = "relu",
                 block_normalization: NormalizationType = "batch",
                 use_skip_connection: bool = True,
                 **kwargs):
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_values,
                              name="rtdl_resnet_regression_head")
        super().__init__(input_dim=input_dim,
                         num_blocks=num_blocks,
                         block_hidden_dim=block_hidden_dim,
                         block_dropout_hidden=block_dropout_hidden,
                         block_dropout_out=block_dropout_out,
                         block_activation_hidden=block_activation_hidden,
                         block_activation_out=block_activation_out,
                         block_normalization=block_normalization,
                         use_skip_connection=use_skip_connection,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_outputs': self.num_outputs,
                       'head_units_values': self.head_units_values,
                       }
                      )
        return config



class BaseConfig:
    """
    Base Config class for all the Config classes.
    It contains all the common functionality.
    """
    @classmethod
    def to_dict(cls):
        attrs = {}
        current_class = cls
        while current_class:
            attrs_temp = vars(cls)
            attrs_temp = {name: value for name, value in attrs_temp.items() if not name.startswith("__")}
            attrs.update(attrs_temp)
            if current_class.__base__.__name__ == "BaseConfig":
                current_class = None
            else:
                current_class = current_class.__base__
        return attrs


class UserModifiableBaseConfig:
    """
    Base Config class for all the Config classes whose instance attributes
    are expected to be modified by the users.
    It contains all the common functionality.
    """
    def to_dict(self):
        attrs = self.__dict__
        attrs = {name: value for name, value in attrs.items() if not name.startswith("__")}
        return attrs



import tensorflow as tf
from tensorflow import keras
from teras.layerflow.models.rtdl_resnet import RTDLResNet
from teras.layers import RTDLResNetBlock
import os


class NODETest(tf.test.TestCase):
    def setUp(self):
        self.inputs = tf.ones((8, 5))
        self.input_dim = 5
        self.resnet_blocks = [RTDLResNetBlock() for _ in range(3)]
        self.head = keras.layers.Dense(1)

    def test_valid_call(self):
        model = RTDLResNet(input_dim=self.input_dim,
                           resnet_blocks=self.resnet_blocks,
                           head=self.head)
        model(self.inputs)

    def test_save_and_load(self):
        model = RTDLResNet(input_dim=self.input_dim,
                           resnet_blocks=self.resnet_blocks,
                           head=self.head)
        save_path = os.path.join(self.get_temp_dir(), "rtdl_resnet_lf.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(self.inputs)
        outputs_reloaded = reloaded_model(self.inputs)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)



import tensorflow as tf
from teras.layers.ft_transformer.ft_cls_token import FTCLSToken


def test_ft_cls_token_valid_call():
    inputs = tf.ones((8, 10, 16))
    cls_token = FTCLSToken(embedding_dim=16)
    outputs = cls_token(inputs)


def test_ft_cls_token_output_shape():
    inputs = (tf.ones((8, 10, 16)))
    cls_token = FTCLSToken(embedding_dim=16)
    outputs = cls_token(inputs)

    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 11
    assert tf.shape(outputs)[2] == 16



import tensorflow as tf
from teras.layers.saint.saint_transformer import SAINTTransformer


def test_saint_transformer_output_shape():
    saint_transformer = SAINTTransformer(data_dim=10,
                                         embedding_dim=32)
    inputs = tf.ones((16, 10, 32), dtype=tf.float32)
    outputs = saint_transformer(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 16
    assert tf.shape(outputs)[1] == 10
    assert tf.shape(outputs)[2] == 32



from teras.layers.common.transformer import FeedForward, Transformer, Encoder
from teras.layers import SAINTNumericalFeatureEmbedding
import tensorflow as tf


def test_feedforward_output_shape():
    feed_forward = FeedForward(embedding_dim=32)
    inputs = tf.ones((128, 16), dtype=tf.float32)
    outputs = feed_forward(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 32


def test_transformer_output_shape():
    transformer = Transformer(embedding_dim=32)
    inputs = tf.ones((128, 16, 32), dtype=tf.float32)
    outputs = transformer(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 16
    assert tf.shape(outputs)[2] == 32


def test_encoder_output_shape():
    encoder = Encoder(embedding_dim=32)
    inputs = tf.ones((128, 16, 32), dtype=tf.float32)
    outputs = encoder(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 16
    assert tf.shape(outputs)[2] == 32




from tensorflow import keras
import numpy as np
import tensorflow as tf
from teras.utils.types import Number


@keras.saving.register_keras_serializable(package="teras.layers")
class CategoricalFeatureEmbedding(keras.layers.Layer):
    """
    CategoricalFeatureEmbedding layer that encodes categorical features into
    categorical feature embeddings.

    Args:
        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this metadata dictionary by calling
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> features_metadata = get_features_metadata_for_embedding(dataset,
                ..                                                          categorical_features,
                ..                                                          numerical_features)
        embedding_dim: ``int``, default 32,
            Dimensionality of the embeddings to which categorical features should be mapped.

        encode: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegrerLookup`` layer.
    """
    def __init__(self,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 encode: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        if not len(features_metadata["categorical"]) > 0:
            raise ValueError("`features_metadata` does not contain any categorical features. "
                             "Either you forgot to pass categorical features names list to the "
                             "`get_features_metadata_for_embedding` or the dataset does not contain "
                             "any categorical features to begin with. \n"
                             "In either case, "
                             "`CategoricalFeatureEmbedding` cannot be called on inputs with no categorical features. ")

        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim
        self.encode = encode

        self.categorical_features_metadata = features_metadata["categorical"]
        self.lookup_tables, self.embedding_layers = self._get_lookup_tables_and_embedding_layers()
        self._num_categorical_features = len(self.categorical_features_metadata)

    def _get_lookup_tables_and_embedding_layers(self):
        """Lookup tables and embedding layers for each categorical feature"""
        lookup_tables = []
        embedding_layers = []
        for feature_name, (feature_idx, vocabulary) in self.categorical_features_metadata.items():
            # If the categorical values in the dataset haven't been encoded in preprocessing step
            # then encode them first. The `encode` parameter lets user specify if we need to encode
            # the categorical values or not.
            if self.encode:
                if isinstance(vocabulary[0], Number):
                    # Lookup Table to map integer values to integer indices
                    lookup = keras.layers.IntegerLookup(vocabulary=vocabulary,
                                                        mask_token=None,
                                                        output_mode="int",
                                                        )
                else:
                    raise TypeError("`CategoricalFeatureEmbedding` layer can only encode values of type int "
                                    f"but received type: {type(vocabulary[0])} for feature {feature_name}.")
                lookup_tables.append(lookup)

            # Embedding layers map the integer representations of categorical values
            # to dense vectors of `embedding_dim` dimensionality,
            # which in fancier lingo are called `embeddings`.
            embedding = keras.layers.Embedding(input_dim=len(vocabulary) + 1,
                                               output_dim=self.embedding_dim)
            embedding_layers.append(embedding)

        return lookup_tables, embedding_layers

    def call(self, inputs):
        # Encode and embedd categorical features
        categorical_feature_embeddings = tf.TensorArray(size=self._num_categorical_features,
                                                        dtype=tf.float32)
        # feature_idx is the overall index of feature in the dataset
        # so it can't be used to retrieve lookup table and embedding layer from list
        # which are both of length equal to number of categorical features - not input dimensions
        # hence the need for current_idx
        current_idx = 0
        # We only embedd the categorical features
        for feature_name, (feature_idx, _) in self.categorical_features_metadata.items():
            feature = tf.expand_dims(inputs[:, feature_idx], 1)
            if self.encode:
                # Convert string input values to integer indices
                lookup = self.lookup_tables[current_idx]
                feature = lookup(feature)
            # Convert index values to embedding representations
            embedding = self.embedding_layers[current_idx]
            feature = embedding(feature)
            categorical_feature_embeddings = categorical_feature_embeddings.write(current_idx, feature)
            current_idx += 1
        categorical_feature_embeddings = categorical_feature_embeddings.stack()
        categorical_feature_embeddings = tf.squeeze(categorical_feature_embeddings, axis=2)
        categorical_feature_embeddings = tf.transpose(categorical_feature_embeddings, perm=[1, 0, 2])
        categorical_feature_embeddings.set_shape((None, self._num_categorical_features, self.embedding_dim))
        return categorical_feature_embeddings

    def get_config(self):
        config = super().get_config()
        new_config = {'features_metadata': self.features_metadata,
                      'embedding_dim': self.embedding_dim,
                      'encode': self.encode}
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        features_metadata = config.pop("features_metadata")
        return cls(features_metadata, **config)



import tensorflow as tf
import pandas as pd
import numpy as np
from sklearn.preprocessing import OrdinalEncoder
from teras.preprocessing.base.base_data_transformer import BaseDataTransformer as _BaseDataTransformer
from teras.utils.types import FeaturesNamesType


class GAINDataTransformer(_BaseDataTransformer):
    """
    GAINDataTransformer class that performs the required transformations
    on the raw dataset required by the GAIN architecture proposed by
    Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        categorical_features: ``List[str]``,
            List of categorical features names in the dataset.
            Categorical features are encoded by ordinal encoder method.
            And then MinMax normalization is applied.

        numerical_features: ``List[str]``,
            List of numerical features names in the dataset.
            Numerical features are encoded using MinMax normalization.

    Example:
        .. code-block:: python

            input_df = pd.DataFrame(np.arange(10, 3).reshape(20, 5),
                                        columns=['A', 'B', 'C'])
            numerical_features = input_df.columns
            data_transformer = DataTransformer(categorical_features=None,
                                               numerical_features=numerical_features)
            transformed_df = data_transformer.transform(input_df)
    """
    def __init__(self,
                 categorical_features: FeaturesNamesType = None,
                 numerical_features: FeaturesNamesType = None
                 ):
        super().__init__()
        self.categorical_features = categorical_features
        self.numerical_features = numerical_features
        self.encoder = OrdinalEncoder()
        self.min_vals = None
        self.max_vals = None

    def fit(self, x,
            **kwargs):
        # We min max normalize the entire dataset regardless of the
        # features types - so we first need to have the ordinally
        # encoded values for the categorical features
        # So we create a temporary copy which contains these
        # encoded values so we can compute the min_vals and max_vals
        # for these features
        x_temp = x.copy()
        if self.categorical_features is not None:
            self.encoder.fit(x[self.categorical_features])
            x_temp[self.categorical_features] = self.encoder.transform(x[self.categorical_features])

        # if self.numerical_features is not None:
        #     self.min_vals = np.nanmin(x[self.numerical_features], axis=0)
        #     self.max_vals = np.nanmax(x[self.numerical_features], axis=0)
        self.min_vals = np.nanmin(x_temp, axis=0)
        self.max_vals = np.nanmax(x_temp, axis=0)
        self.fitted = True

    def transform(self,
                  x: pd.DataFrame,
                  **kwargs):
        """
        Transforms the data (applying normalization etc)
        and returns a tensorflow dataset.
        It also stores the metadata of features
        that is used in the reverse transformation step.

        Args:
            x: Data to transform. Must be a pandas DataFrame.

        Returns:
            Transformed data.
        """
        if not self.fitted:
            raise RuntimeError("You haven't yet fitted the DataTransformer. "
                               "You must call the `fit` method before you can call the "
                               "`transform` method. ")
        if not isinstance(x, pd.DataFrame):
            raise ValueError("Only pandas dataframe is supported by DataTransformation class."
                             f" But data of type {type(x)} was passed. "
                             f"Please convert it to pandas dataframe before passing.")

        self.ordered_features_names_all = x.columns

        if self.categorical_features is not None:
            x[self.categorical_features] = self.encoder.transform(x[self.categorical_features])

        # we don't need to check if there are numerical features,
        # because we apply this "numerical" MinMax transformation to the whole
        # dataset, even to the categorical features we converted using ordinal encoder
        x = x.values
        x = (x - self.min_vals) / self.max_vals
        x = x.astype(np.float)

        # if return_dataframe:
        x = pd.DataFrame(x, columns=self.ordered_features_names_all)
        return x

    def reverse_transform(self,
                          x,
                          **kwargs):
        """
        Reverse Transforms the transformed data.
        The `min_vals` and `max_vals` values learnt in the `transform` step
        are used for reverse transformation of numerical features.

        Args:
            x: Transformed Data.

        Returns:
            Data in its original format and scale.
        """
        if not isinstance(x, pd.DataFrame):
            x = pd.DataFrame(x, columns=self.ordered_features_names_all)

        if self.min_vals is None and self.max_vals is None:
            raise ValueError("The values for `min_vals` and `max_vals` do not exist. "
                             "This implies that you haven't yet used the `transform` method "
                             "to transform numerical features. ")

        # Since we apply the numerical minmax transformation to the whole dataset
        # even the ordinal encoded categorical features, so we first reverse
        # transform the numerical transformations on the whole dataset and then
        # we reverse categorical transformations if any.
        if self.numerical_features is not None:
            x = (x * self.max_vals) + self.min_vals
        if self.categorical_features is not None:
            x[self.categorical_features] = self.encoder.inverse_transform(x[self.categorical_features])

        # if not return_dataframe:
        #     x = x.values
        return x


class GAINDataSampler:
    """
    GAINDataSampler class that prepares the transformed data in the format
    that is expected and digestible by the GAIN architecture proposed by
    Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        batch_size: ``int``, default 512,
            Batch size to use for dataset.

        shuffle: ``bool``, default True,
            Whether to shuffle the data.

        random_seed: ``int``, default None,
            Random seed to use when shuffling.

    Example:
        ```python
        input_data = np.random.arange(100).reshape(20, 5)
        data_sampler = DataSampler()
        dataset = data_sampler.get_dataset(input_data)
        ```
    """
    def __init__(self,
                 batch_size=512,
                 shuffle=True,
                 random_seed=None):
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.random_seed = random_seed
        self.num_samples = None
        self.data_dim = None

        if self.random_seed:
            np.random.seed(self.random_seed)

    def get_dataset(self, x_transformed):
        """
        Args:
            x_transformed: ``np.ndarray`` or ``pd.DataFrame``,
                Transformed dataset.
        Returns:
             A tensorflow dataset, that generates batches of
            size `batch_size` which is the argument passed during
            DataSampler instantiation. The dataset instance then can
            be readily passed to the GAIN fit method without requiring
            any further processing.
        """
        self.num_samples, self.data_dim = x_transformed.shape

        dataset = (tf.data.Dataset
                   .from_generator(self.generator,
                                   output_signature=(
                                       tf.TensorSpec(shape=(None, self.data_dim), dtype=tf.float32, name="x_generator"),
                                       tf.TensorSpec(shape=(None, self.data_dim), dtype=tf.float32, name="x_discriminator"),
                                   ),
                                   args=[x_transformed]
                                   )
                   )
        return dataset

    def generator(self, x):
        """
        Generator function that is used to create tensorflow dataset.
        It applies any manipulations that are required and then generates
        batches in the format required by the architecture.

        Args:
            x: Dataset to be converted into tensorflow dataset.
        """
        steps_per_epoch = self.num_samples // self.batch_size
        is_n_divisible_by_batch_size = self.num_samples % self.batch_size == 0
        steps_per_epoch += 1 if not is_n_divisible_by_batch_size else 0

        # since we need to draw a batch of data separately for generator and discriminator
        # so we generate idx separately for generator and discriminator
        # and shuffle them if necessary
        generator_idx = np.arange(self.num_samples)
        discriminator_idx = np.arange(self.num_samples)
        if self.shuffle:
            generator_idx = np.random.permutation(generator_idx)
            discriminator_idx = np.random.permutation(discriminator_idx)

        from_index = 0
        to_index = self.batch_size
        for i in range(steps_per_epoch):
            # One the last batch when the number of samples isn't divisible
            # by batch size, then trying to extract x[from_index: to_index] will
            # result in error since to_index will go out of bounds.
            # So to avoid this, we add the following check
            # and modify the to_index value accordingly.
            if i == (steps_per_epoch - 1) and not is_n_divisible_by_batch_size:
                x_generator = x[generator_idx[from_index: None]]
                x_discriminator = x[discriminator_idx[from_index: None]]
            else:
                x_generator = x[generator_idx[from_index: to_index]]
                x_discriminator = x[discriminator_idx[from_index: to_index]]
                from_index += self.batch_size
                to_index += self.batch_size

            yield x_generator, x_discriminator



import tensorflow as tf
from tensorflow import keras
import tensorflow_probability as tfp


@keras.saving.register_keras_serializable(package="teras.layers.regularization")
class CutMix(keras.layers.Layer):
    """
    CutMix is a regularization layer proposed by Sangdoo Yun et al.
    in the paper,
    CutMix: Regularization Strategy to Train Strong Classifiers
    with Localizable Features

    It was originally proposed for image data but here it has been
    adapted for Tabular data.

    Args:
        probs: `float`, default 0.3
            CutMix probability which is used in generation of mask that is used
            to mix samples together.
    """
    def __init__(self,
                 probs: float = 0.3,
                 **kwargs):
        super().__init__(**kwargs)
        self.probs = probs
        self.mask_generator = tfp.distributions.Binomial(total_count=1,
                                                         probs=self.probs)

    def call(self, inputs):
        # Generate mask for CutMix mixing
        mask_cutmix = self.mask_generator.sample(sample_shape=tf.shape(inputs))

        # For each data sample select a partner to mix it with at random.
        # To efficiently achieve this, we can just shuffle the data
        random_partners = tf.random.shuffle(inputs)

        # Apply cutmix formula
        inputs_cutmixed = (inputs * mask_cutmix) + (random_partners * (1 - mask_cutmix))
        return inputs_cutmixed

    def get_config(self):
        config = super().get_config()
        config.update({'probs': self.probs,
                       })
        return config



import tensorflow as tf
from tensorflow import keras
from teras.models.ctgan import CTGANGenerator, CTGANDiscriminator, CTGAN
from teras.preprocessing.ctgan import CTGANDataTransformer, CTGANDataSampler
from teras.utils import inject_missing_values, generate_fake_gemstone_data
import os


class CTGANGeneratorTest(tf.test.TestCase):
    def setUp(self):
        fake_gem_df = generate_fake_gemstone_data(num_samples=32)
        num_cols = ["depth", "table"]
        cat_cols = ["cut", "color", "clarity"]
        self.data_transformer = CTGANDataTransformer(numerical_features=num_cols,
                                                     categorical_features=cat_cols)
        x_transformed = self.data_transformer.transform(fake_gem_df)

        self.data_sampler = CTGANDataSampler(batch_size=8,
                                             categorical_features=cat_cols,
                                             metadata=self.data_transformer.get_metadata())
        self.dataset = self.data_sampler.get_dataset(x_transformed=x_transformed,
                                                     x_original=fake_gem_df)
        self.metadata = self.data_transformer.get_metadata()

    def test_valid_call(self):
        z = tf.random.uniform((8, 16))
        generator = CTGANGenerator(data_dim=5,
                                   metadata=self.metadata)
        outputs = generator(z)

    def test_output_shape(self):
        z = tf.random.uniform((8, 16))
        generator = CTGANGenerator(data_dim=5,
                                   metadata=self.metadata)
        outputs = generator(z)
        self.assertAllEqual(tf.shape(outputs), (8, 5))

    def test_save_and_load(self):
        z = tf.random.uniform((8, 16))
        generator = CTGANGenerator(data_dim=5,
                                   metadata=self.metadata)
        save_path = os.path.join(self.get_temp_dir(), "ctgan_generator.keras")
        generator.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = generator(z)
        outputs_reloaded = reloaded_model(z)
        # self.assertAllClose(outputs_original, outputs_reloaded)


class CTGANDiscriminatorTest(tf.test.TestCase):
    def setUp(self):
        self.inputs = tf.ones((8, 5))

    def test_valid_call(self):
        discriminator = CTGANDiscriminator()
        outputs = discriminator(self.inputs)

    def test_output_shape(self):
        discriminator = CTGANDiscriminator()
        outputs = discriminator(self.inputs)
        self.assertAllEqual(tf.shape(outputs), (1, 1))

    def test_save_and_load(self):
        discriminator = CTGANDiscriminator()
        save_path = os.path.join(self.get_temp_dir(), "ctgan_discriminator.keras")
        discriminator.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = discriminator(self.inputs)
        outputs_reloaded = reloaded_model(self.inputs)
        # self.assertAllClose(outputs_original, outputs_reloaded)


class CTGANTest(tf.test.TestCase):
    def setUp(self):
        fake_gem_df = generate_fake_gemstone_data(num_samples=32)
        num_cols = ["depth", "table"]
        cat_cols = ["cut", "color", "clarity"]
        self.data_transformer = CTGANDataTransformer(numerical_features=num_cols,
                                                     categorical_features=cat_cols)
        x_transformed = self.data_transformer.transform(fake_gem_df)

        self.data_sampler = CTGANDataSampler(batch_size=8,
                                             categorical_features=cat_cols,
                                             metadata=self.data_transformer.get_metadata())
        self.dataset = self.data_sampler.get_dataset(x_transformed=x_transformed,
                                                     x_original=fake_gem_df)
        self.metadata = self.data_transformer.get_metadata()
        self.data_dim = self.data_sampler.data_dim
        self.ctgan = CTGAN(input_dim=self.data_dim,
                           metadata=self.metadata)

    def test_valid_call(self):
        # CTGAN's call method is just calls the generator's call method
        z = tf.random.normal((8, 16))
        ctgan = CTGAN(input_dim=self.data_dim,
                      metadata=self.metadata)
        outputs = ctgan(z)

    def test_output_shape(self):
        # CTGAN's call method is just calls the generator's call method
        z = tf.random.normal((8, 16))
        ctgan = CTGAN(input_dim=self.data_dim,
                      metadata=self.metadata)
        outputs = ctgan(z)
        self.assertAllEqual(tf.shape(outputs), (8, self.data_dim))

    def test_save_and_load(self):
        z = tf.random.normal((8, 16))
        ctgan = CTGAN(input_dim=self.data_dim,
                      metadata=self.metadata)
        save_path = os.path.join(self.get_temp_dir(), "ctgan.keras")
        ctgan.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = ctgan(z)
        outputs_reloaded = reloaded_model(z)
        # self.assertAllClose(outputs_original, outputs_reloaded)

    def test_fit(self):
        self.ctgan.compile()
        self.ctgan.fit(self.dataset)

    def test_generate(self):
        generated_data = self.ctgan.generate(num_samples=16,
                                             data_sampler=self.data_sampler,
                                             data_transformer=self.data_transformer,
                                             reverse_transform=True)
        assert len(generated_data) == 16



import tensorflow as tf
from tensorflow.keras import losses


def discriminator_loss(discriminator_pred, mask):
    """
    Discriminator loss based on the architecture and implementation
    of GAIN as proposed and implemented by
    Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    It computes the cross entropy loss of discriminator's predictions.

    Reference(s):
        https://arxiv.org/abs/1806.02920
        https://github.com/jsyoon0823/GAIN

    Args:
        discriminator_pred: Discriminator's predictions
        mask: Mask generated from inputs.

    Returns:
        Cross entropy loss for discriminator's predictions.
    """
    loss = -tf.reduce_mean(mask * tf.math.log(discriminator_pred + 1e-8)
                           + (1 - mask) * tf.math.log(1. - discriminator_pred + 1e-8))
    return loss


def generator_loss(real_samples=None,
                   generated_samples=None,
                   discriminator_pred=None,
                   mask=None,
                   alpha=None):
    """
    Generator loss based on the architecture and implementation
    of GAIN as proposed and implemented by
    Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    This loss is made up of two components,
    cross_entropy_loss: That is computed for discriminator's predictions.
    mse_loss: That is computed between the masked original and generated samples
    These losses are combined with `alpha` determining the weight of mse loss
    Precisely, `loss` = `cross_entropy_loss` + `alpha` * `mse_loss`

    Reference(s):
        https://arxiv.org/abs/1806.02920
        https://github.com/jsyoon0823/GAIN

    Args:
        real_samples: Real samples drawn from the dataset
        generated_samples: Samples generated by Generator
        discriminator_pred: Predictions output by Discriminator
        mask: Mask generated from the inputs
        alpha: The hyperparameter alpha used to control the
            weightage during addition of two losses.

    Returns:
        Generator loss made up cross entropy and MSE loss.
    """
    cross_entropy_loss = -tf.reduce_mean((1 - mask) * tf.math.log(discriminator_pred + 1e-8))
    mse_loss = losses.MSE(y_true=(mask * real_samples),
                            y_pred=(mask * generated_samples))
    loss = cross_entropy_loss + alpha * mse_loss
    return loss



import tensorflow as tf
from tensorflow import keras
import pandas as pd
import numpy as np
from teras.models.saint import SAINTClassifier, SAINTRegressor, SAINT, SAINTPretrainer
from teras.utils import get_features_metadata_for_embedding
import os


class SAINTClassifierTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.data_batch = data.values

    def test_valid_call(self):
        model = SAINTClassifier(features_metadata=self.features_metadata,
                                input_dim=4)
        model(self.data_batch)

    def test_save_and_load(self):
        model = SAINTClassifier(features_metadata=self.features_metadata,
                                input_dim=4)
        save_path = os.path.join(self.get_temp_dir(), "saint_classifier.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)


class SAINTRegressorTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.data_batch = data.values

    def test_valid_call(self):
        model = SAINTRegressor(features_metadata=self.features_metadata,
                               input_dim=4)
        model(self.data_batch)

    def test_save_and_load(self):
        model = SAINTRegressor(features_metadata=self.features_metadata,
                               input_dim=4)
        save_path = os.path.join(self.get_temp_dir(), "saint_regressor.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)


class SAINTPretrainerTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.data_batch = data.values

    def test_valid_call(self):
        base_model = SAINT(features_metadata=self.features_metadata,
                           input_dim=4)
        pretrainer = SAINTPretrainer(model=base_model,
                                     features_metadata=self.features_metadata)
        pretrainer(self.data_batch)

    def test_save_and_load(self):
        base_model = SAINT(features_metadata=self.features_metadata,
                           input_dim=4)
        pretrainer = SAINTPretrainer(model=base_model,
                                     features_metadata=self.features_metadata)
        save_path = os.path.join(self.get_temp_dir(), "saint_pretrainer.keras")
        pretrainer.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = pretrainer(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)



import tensorflow as tf
from tensorflow import keras
import tensorflow_probability as tfp


@keras.saving.register_keras_serializable(package="teras.layers.regularization")
class MixUp(keras.layers.Layer):
    """
    MixUp is a regularization layer proposed by Hongyi Zhang et al.
    in the paper,
    mixup: BEYOND EMPIRICAL RISK MINIMIZATION

    It was originally proposed for image data but here it has been
    adapted for Tabular data.

    Reference(s):
        https://arxiv.org/abs/1710.09412

    Args:
        alpha: `float`, default 1.0,
            Parameter for the Beta distribution to sample `lambda_`
            from which is used to interpolate samples.
    """
    def __init__(self,
                 alpha: float = 1.,
                 **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.lambda_generator = tfp.distributions.Beta(self.alpha, self.alpha)

    def call(self, inputs):
        # Sample lambda_
        lambda_ = tf.squeeze(self.lambda_generator.sample(1))
        # For each data sample select a partner to mix it with at random.
        # To efficiently achieve this, we can just shuffle the data
        random_partners = tf.random.shuffle(inputs)

        # SIDE NOTE:
        # We could make the shuffling more memory efficient by just shuffling the indices
        # but the problem with that approach is we will have to check and handle the
        # dictionary type data differently -- making it lot more complex and perhaps
        # worse in terms of efficiency

        inputs_mixedup = (lambda_ * inputs) + (1 - lambda_) * random_partners
        return inputs_mixedup

    def get_config(self):
        config = super().get_config()
        config.update({'alpha': self.alpha,
                       })
        return config



import tensorflow as tf
from tensorflow.keras import losses
from teras.losses.gain import (generator_loss as generator_pretraining_loss,
                               discriminator_loss)


# The generator pretraining loss is the same as generator_loss
# in the GAIN model because in PC-GAIN we use a GAIN model for
# pretraining.

# In PC-GAIN, the discriminator loss is also exactly the same
# as GAIN's discriminator loss

def generator_loss(real_samples=None,
                   generated_samples=None,
                   discriminator_pred=None,
                   mask=None,
                   alpha=200,
                   beta=100,
                   classifier_pred=None):
    """
    Generator loss used during the main training phase (post-pretraining)
    of PC-GAIN architecture.
    It is similar to the generator loss used in the pretraining stage
    except for an additional `Information Entropy Loss` that is calculated
    for the Classifier's predictions and weighted by the `beta` parameter.

    Args:
        real_samples: Samples drawn from the input dataset.
        generated_samples: Samples generated by the Generator.
        discriminator_pred: Predictions made by Discriminator,
            on the real and generated samples combined.
        mask: Mask created from the raw data batch with missing values,
            it helps identify which values were original missing and
            have been imputed and filter out generated values that were
            not missing in the batch.
            Important to remember that, the Generator generates all values
            for each sample, but we only want values in place of those that
            were originally missing. So we need some way for this identification
            and filteration. This is where this mask comes in.
            In the case of loss computation, it helps apply and combined relevant
            functions and values.
        alpha: Hyperparamter that controls how much the MSE loss contributes to
            the overall loss. It is directly proportional to the MSE loss.
            Important to keep in mind that the overall loss is made up of three losses,
            `Cross Entropy loss`, `MSE loss` and `Info entropy loss`
        beta: Hyperparamter that controls how much the Info Entropy loss contributes to
            the overall loss. It is directly proportional to the Info Entropy loss.
        classifier_pred: Softmax probabilities predicted by the classifier for the
            given batch of fused together real and generated samples.
    """
    cross_entropy_loss = -tf.reduce_mean((1 - mask) * tf.math.log(discriminator_pred + 1e-8))
    mse_loss = losses.MSE(y_true=(mask * real_samples),
                                  y_pred=(mask * generated_samples))
    info_entropy_loss = -tf.reduce_mean(classifier_pred * tf.math.log(classifier_pred + 1e-8))
    loss = cross_entropy_loss + (alpha * mse_loss) + (beta * info_entropy_loss)
    return loss



# Copyright 2023 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Common layers
# -> Common Head layers
from teras.layers.common.head import (ClassificationHead,
                                      RegressionHead)
# -> Common Transformer layers
from teras.layers.common.transformer import (FeedForward,
                                             Transformer,
                                             Encoder)


# Embedding layers
from teras.layers.categorical_feature_embedding import CategoricalFeatureEmbedding


# NODE layers
from teras.layers.node.odst import ObliviousDecisionTree
from teras.layers.node.node_feature_selector import NodeFeatureSelector


# TabNet layers
from teras.layers.tabnet.tabnet_attentive_transformer import TabNetAttentiveTransformer
from teras.layers.tabnet.tabnet_feature_transformer_block import TabNetFeatureTransformerBlock
from teras.layers.tabnet.tabnet_feature_transformer import TabNetFeatureTransformer
from teras.layers.tabnet.tabnet_encoder import TabNetEncoder
from teras.layers.tabnet.tabnet_decoder import TabNetDecoder


# TabTransformer layers
from teras.layers.tabtransformer.tabtransformer_column_embedding import TabTransformerColumnEmbedding


# DNFNet layers
from teras.layers.dnfnet.dnfnet_feature_selection import DNFNetFeatureSelection
from teras.layers.dnfnet.dnfnet_localization import DNFNetLocalization
from teras.layers.dnfnet.dnnf import DNNF


# SAINT layers
from teras.layers.saint.saint_numerical_feature_embedding import SAINTNumericalFeatureEmbedding
from teras.layers.saint.multi_head_inter_sample_attention import MultiHeadInterSampleAttention
from teras.layers.saint.saint_transformer import SAINTTransformer
from teras.layers.saint.saint_encoder import SAINTEncoder
from teras.layers.saint.saint_projection_head import SAINTProjectionHead
from teras.layers.saint.saint_reconstruction_head import (SAINTReconstructionHeadBlock,
                                                          SAINTReconstructionHead)


# RTDL ResNet layers
from teras.layers.rtdl_resnet.rtdl_resnet_block import RTDLResNetBlock


# RTDL FTTransformer layers
from teras.layers.ft_transformer.ft_numerical_feature_embedding import FTNumericalFeatureEmbedding
from teras.layers.ft_transformer.ft_cls_token import FTCLSToken


# VIME layers
from teras.layers.vime.vime_mask_estimator import VimeMaskEstimator
from teras.layers.vime.vime_feature_estimator import VimeFeatureEstimator
from teras.layers.vime.vime_encoder import VimeEncoder
from teras.layers.vime.vime_predictor import VimePredictor
from teras.layers.vime.vime_mask_generation_and_corruption import VimeMaskGenerationAndCorruption


# On Embeddings for Numerical Features (paper) layers
from teras.layers.preprocessing.periodic_embedding import (PeriodicEmbedding)


# Regularization layers
from teras.layers.regularization.mixup import MixUp
from teras.layers.regularization.cutmix import CutMix


# Encoding layers
from teras.layers.preprocessing.label_encoding import LabelEncoding


# GAIN layers
from teras.layers.gain.gain_generator_block import GAINGeneratorBlock
from teras.layers.gain.gain_discriminator_block import GAINDiscriminatorBlock


# CTGAN layers
from teras.layers.ctgan.ctgan_generator_block import CTGANGeneratorBlock
from teras.layers.ctgan.ctgan_discriminator_block import CTGANDiscriminatorBlock


# Normalization layers
from teras.layers.numerical_feature_normalization import NumericalFeatureNormalization



import tensorflow as tf
from teras.layers.tabnet.tabnet_encoder import TabNetEncoder


def test_tabnet_encoder_valid_call():
    encoder = TabNetEncoder(data_dim=10,
                            feature_transformer_dim=32,
                            decision_step_output_dim=32)
    inputs = tf.ones((8, 10))
    outputs = encoder(inputs)


def test_tabnet_encoder_output_shape():
    encoder = TabNetEncoder(data_dim=10,
                            feature_transformer_dim=32,
                            decision_step_output_dim=32)
    inputs = tf.ones((8, 10))
    outputs = encoder(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 32
